# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Public API for tf.raw_ops namespace.
"""

from __future__ import print_function as _print_function

from tensorflow.python import Abs
from tensorflow.python import AccumulateNV2
from tensorflow.python import AccumulatorApplyGradient
from tensorflow.python import AccumulatorNumAccumulated
from tensorflow.python import AccumulatorSetGlobalStep
from tensorflow.python import AccumulatorTakeGradient
from tensorflow.python import Acos
from tensorflow.python import Acosh
from tensorflow.python import Add
from tensorflow.python import AddManySparseToTensorsMap
from tensorflow.python import AddN
from tensorflow.python import AddSparseToTensorsMap
from tensorflow.python import AddV2
from tensorflow.python import All
from tensorflow.python import Angle
from tensorflow.python import Any
from tensorflow.python import ApproximateEqual
from tensorflow.python import ArgMax
from tensorflow.python import ArgMin
from tensorflow.python import AsString
from tensorflow.python import Asin
from tensorflow.python import Asinh
from tensorflow.python import Atan
from tensorflow.python import Atan2
from tensorflow.python import Atanh
from tensorflow.python import BarrierClose
from tensorflow.python import BarrierIncompleteSize
from tensorflow.python import BarrierInsertMany
from tensorflow.python import BarrierReadySize
from tensorflow.python import BarrierTakeMany
from tensorflow.python import Batch
from tensorflow.python import BatchCholesky
from tensorflow.python import BatchCholeskyGrad
from tensorflow.python import BatchFunction
from tensorflow.python import BatchMatMul
from tensorflow.python import BatchMatMulV2
from tensorflow.python import BatchMatrixBandPart
from tensorflow.python import BatchMatrixDeterminant
from tensorflow.python import BatchMatrixDiag
from tensorflow.python import BatchMatrixDiagPart
from tensorflow.python import BatchMatrixInverse
from tensorflow.python import BatchMatrixSetDiag
from tensorflow.python import BatchMatrixSolve
from tensorflow.python import BatchMatrixSolveLs
from tensorflow.python import BatchMatrixTriangularSolve
from tensorflow.python import BatchSelfAdjointEig
from tensorflow.python import BatchSelfAdjointEigV2
from tensorflow.python import BatchSvd
from tensorflow.python import BatchToSpace
from tensorflow.python import BatchToSpaceND
from tensorflow.python import BesselI0e
from tensorflow.python import BesselI1e
from tensorflow.python import Betainc
from tensorflow.python import Bincount
from tensorflow.python import Bitcast
from tensorflow.python import BroadcastArgs
from tensorflow.python import BroadcastGradientArgs
from tensorflow.python import BroadcastTo
from tensorflow.python import Bucketize
from tensorflow.python import Cast
from tensorflow.python import Ceil
from tensorflow.python import CheckNumerics
from tensorflow.python import Cholesky
from tensorflow.python import CholeskyGrad
from tensorflow.python import ClipByValue
from tensorflow.python import CompareAndBitpack
from tensorflow.python import Complex
from tensorflow.python import ComplexAbs
from tensorflow.python import Concat
from tensorflow.python import ConcatOffset
from tensorflow.python import ConcatV2
from tensorflow.python import Conj
from tensorflow.python import ConjugateTranspose
from tensorflow.python import Const
from tensorflow.python import Cos
from tensorflow.python import Cosh
from tensorflow.python import Cross
from tensorflow.python import Cumprod
from tensorflow.python import Cumsum
from tensorflow.python import DebugGradientIdentity
from tensorflow.python import DebugGradientRefIdentity
from tensorflow.python import DecodeBase64
from tensorflow.python import DecodeCSV
from tensorflow.python import DecodeCompressed
from tensorflow.python import DecodeJSONExample
from tensorflow.python import DecodePaddedRaw
from tensorflow.python import DecodeRaw
from tensorflow.python import DeepCopy
from tensorflow.python import DeleteSessionTensor
from tensorflow.python import DepthToSpace
from tensorflow.python import Dequantize
from tensorflow.python import DeserializeManySparse
from tensorflow.python import DeserializeSparse
from tensorflow.python import Diag
from tensorflow.python import DiagPart
from tensorflow.python import Digamma
from tensorflow.python import Div
from tensorflow.python import DivNoNan
from tensorflow.python import DynamicPartition
from tensorflow.python import DynamicStitch
from tensorflow.python import EditDistance
from tensorflow.python import Empty
from tensorflow.python import EncodeBase64
from tensorflow.python import EnsureShape
from tensorflow.python import Equal
from tensorflow.python import Erf
from tensorflow.python import Erfc
from tensorflow.python import EuclideanNorm
from tensorflow.python import Exp
from tensorflow.python import ExpandDims
from tensorflow.python import Expm1
from tensorflow.python import ExtractImagePatches
from tensorflow.python import ExtractVolumePatches
from tensorflow.python import FIFOQueueV2
from tensorflow.python import FakeQuantWithMinMaxArgs
from tensorflow.python import FakeQuantWithMinMaxArgsGradient
from tensorflow.python import FakeQuantWithMinMaxVars
from tensorflow.python import FakeQuantWithMinMaxVarsGradient
from tensorflow.python import FakeQuantWithMinMaxVarsPerChannel
from tensorflow.python import FakeQuantWithMinMaxVarsPerChannelGradient
from tensorflow.python import FakeQueue
from tensorflow.python import Fill
from tensorflow.python import Fingerprint
from tensorflow.python import FixedLengthRecordReaderV2
from tensorflow.python import Floor
from tensorflow.python import FloorDiv
from tensorflow.python import FloorMod
from tensorflow.python import Gather
from tensorflow.python import GatherNd
from tensorflow.python import GatherV2
from tensorflow.python import GetSessionHandle
from tensorflow.python import GetSessionHandleV2
from tensorflow.python import GetSessionTensor
from tensorflow.python import Greater
from tensorflow.python import GreaterEqual
from tensorflow.python import GuaranteeConst
from tensorflow.python import HistogramFixedWidth
from tensorflow.python import IdentityN
from tensorflow.python import IdentityReaderV2
from tensorflow.python import Igamma
from tensorflow.python import IgammaGradA
from tensorflow.python import Igammac
from tensorflow.python import Imag
from tensorflow.python import ImmutableConst
from tensorflow.python import InplaceAdd
from tensorflow.python import InplaceSub
from tensorflow.python import InplaceUpdate
from tensorflow.python import Inv
from tensorflow.python import InvGrad
from tensorflow.python import InvertPermutation
from tensorflow.python import IsFinite
from tensorflow.python import IsInf
from tensorflow.python import IsNan
from tensorflow.python import Less
from tensorflow.python import LessEqual
from tensorflow.python import Lgamma
from tensorflow.python import LinSpace
from tensorflow.python import ListDiff
from tensorflow.python import Log
from tensorflow.python import Log1p
from tensorflow.python import LogMatrixDeterminant
from tensorflow.python import LogicalAnd
from tensorflow.python import LogicalNot
from tensorflow.python import LogicalOr
from tensorflow.python import LowerBound
from tensorflow.python import Lu
from tensorflow.python import MapClear
from tensorflow.python import MapIncompleteSize
from tensorflow.python import MapPeek
from tensorflow.python import MapSize
from tensorflow.python import MapStage
from tensorflow.python import MapUnstage
from tensorflow.python import MapUnstageNoKey
from tensorflow.python import MatMul
from tensorflow.python import MatchingFiles
from tensorflow.python import MatrixBandPart
from tensorflow.python import MatrixDeterminant
from tensorflow.python import MatrixDiag
from tensorflow.python import MatrixDiagPart
from tensorflow.python import MatrixExponential
from tensorflow.python import MatrixInverse
from tensorflow.python import MatrixLogarithm
from tensorflow.python import MatrixSetDiag
from tensorflow.python import MatrixSolve
from tensorflow.python import MatrixSolveLs
from tensorflow.python import MatrixSquareRoot
from tensorflow.python import MatrixTriangularSolve
from tensorflow.python import Max
from tensorflow.python import Maximum
from tensorflow.python import Mean
from tensorflow.python import MergeV2Checkpoints
from tensorflow.python import Min
from tensorflow.python import Minimum
from tensorflow.python import MirrorPad
from tensorflow.python import MirrorPadGrad
from tensorflow.python import Mod
from tensorflow.python import Mul
from tensorflow.python import MulNoNan
from tensorflow.python import Multinomial
from tensorflow.python import Neg
from tensorflow.python import NextAfter
from tensorflow.python import NotEqual
from tensorflow.python import OneHot
from tensorflow.python import OnesLike
from tensorflow.python import OrderedMapClear
from tensorflow.python import OrderedMapIncompleteSize
from tensorflow.python import OrderedMapPeek
from tensorflow.python import OrderedMapSize
from tensorflow.python import OrderedMapStage
from tensorflow.python import OrderedMapUnstage
from tensorflow.python import OrderedMapUnstageNoKey
from tensorflow.python import Pack
from tensorflow.python import Pad
from tensorflow.python import PadV2
from tensorflow.python import PaddingFIFOQueueV2
from tensorflow.python import ParallelConcat
from tensorflow.python import ParallelDynamicStitch
from tensorflow.python import ParameterizedTruncatedNormal
from tensorflow.python import ParseExample
from tensorflow.python import ParseSequenceExample
from tensorflow.python import ParseSingleExample
from tensorflow.python import ParseSingleSequenceExample
from tensorflow.python import ParseTensor
from tensorflow.python import Placeholder
from tensorflow.python import PlaceholderV2
from tensorflow.python import PlaceholderWithDefault
from tensorflow.python import Polygamma
from tensorflow.python import Pow
from tensorflow.python import PreventGradient
from tensorflow.python import PriorityQueueV2
from tensorflow.python import Prod
from tensorflow.python import Qr
from tensorflow.python import QuantizeAndDequantize
from tensorflow.python import QuantizeAndDequantizeV2
from tensorflow.python import QuantizeAndDequantizeV3
from tensorflow.python import QuantizeDownAndShrinkRange
from tensorflow.python import QuantizeV2
from tensorflow.python import QuantizedAdd
from tensorflow.python import QuantizedConcat
from tensorflow.python import QuantizedInstanceNorm
from tensorflow.python import QuantizedMatMul
from tensorflow.python import QuantizedMul
from tensorflow.python import QuantizedReshape
from tensorflow.python import QueueClose
from tensorflow.python import QueueCloseV2
from tensorflow.python import QueueDequeue
from tensorflow.python import QueueDequeueMany
from tensorflow.python import QueueDequeueManyV2
from tensorflow.python import QueueDequeueUpTo
from tensorflow.python import QueueDequeueUpToV2
from tensorflow.python import QueueDequeueV2
from tensorflow.python import QueueEnqueue
from tensorflow.python import QueueEnqueueMany
from tensorflow.python import QueueEnqueueManyV2
from tensorflow.python import QueueEnqueueV2
from tensorflow.python import QueueIsClosed
from tensorflow.python import QueueIsClosedV2
from tensorflow.python import QueueSize
from tensorflow.python import QueueSizeV2
from tensorflow.python import RandomGamma
from tensorflow.python import RandomGammaGrad
from tensorflow.python import RandomPoisson
from tensorflow.python import RandomPoissonV2
from tensorflow.python import RandomShuffle
from tensorflow.python import RandomShuffleQueueV2
from tensorflow.python import RandomStandardNormal
from tensorflow.python import RandomUniform
from tensorflow.python import RandomUniformInt
from tensorflow.python import Range
from tensorflow.python import Rank
from tensorflow.python import ReadFile
from tensorflow.python import ReaderNumRecordsProduced
from tensorflow.python import ReaderNumRecordsProducedV2
from tensorflow.python import ReaderNumWorkUnitsCompleted
from tensorflow.python import ReaderNumWorkUnitsCompletedV2
from tensorflow.python import ReaderRead
from tensorflow.python import ReaderReadUpTo
from tensorflow.python import ReaderReadUpToV2
from tensorflow.python import ReaderReadV2
from tensorflow.python import ReaderReset
from tensorflow.python import ReaderResetV2
from tensorflow.python import ReaderRestoreState
from tensorflow.python import ReaderRestoreStateV2
from tensorflow.python import ReaderSerializeState
from tensorflow.python import ReaderSerializeStateV2
from tensorflow.python import Real
from tensorflow.python import RealDiv
from tensorflow.python import Reciprocal
from tensorflow.python import ReciprocalGrad
from tensorflow.python import ReduceJoin
from tensorflow.python import RefIdentity
from tensorflow.python import RegexFullMatch
from tensorflow.python import RegexReplace
from tensorflow.python import RequantizationRange
from tensorflow.python import RequantizationRangePerChannel
from tensorflow.python import Requantize
from tensorflow.python import RequantizePerChannel
from tensorflow.python import Reshape
from tensorflow.python import ResourceStridedSliceAssign
from tensorflow.python import Restore
from tensorflow.python import RestoreSlice
from tensorflow.python import RestoreV2
from tensorflow.python import Reverse
from tensorflow.python import ReverseSequence
from tensorflow.python import ReverseV2
from tensorflow.python import Rint
from tensorflow.python import Round
from tensorflow.python import Rsqrt
from tensorflow.python import RsqrtGrad
from tensorflow.python import Save
from tensorflow.python import SaveSlices
from tensorflow.python import SaveV2
from tensorflow.python import ScatterNd
from tensorflow.python import ScatterNdNonAliasingAdd
from tensorflow.python import SegmentMax
from tensorflow.python import SegmentMean
from tensorflow.python import SegmentMin
from tensorflow.python import SegmentProd
from tensorflow.python import SegmentSum
from tensorflow.python import Select
from tensorflow.python import SelectV2
from tensorflow.python import SelfAdjointEig
from tensorflow.python import SelfAdjointEigV2
from tensorflow.python import SerializeManySparse
from tensorflow.python import SerializeSparse
from tensorflow.python import SerializeTensor
from tensorflow.python import Shape
from tensorflow.python import ShapeN
from tensorflow.python import ShardedFilename
from tensorflow.python import ShardedFilespec
from tensorflow.python import Sigmoid
from tensorflow.python import SigmoidGrad
from tensorflow.python import Sign
from tensorflow.python import Sin
from tensorflow.python import Sinh
from tensorflow.python import Size
from tensorflow.python import Slice
from tensorflow.python import Snapshot
from tensorflow.python import SpaceToBatch
from tensorflow.python import SpaceToBatchND
from tensorflow.python import SpaceToDepth
from tensorflow.python import SparseAccumulatorApplyGradient
from tensorflow.python import SparseAccumulatorTakeGradient
from tensorflow.python import SparseAdd
from tensorflow.python import SparseAddGrad
from tensorflow.python import SparseConcat
from tensorflow.python import SparseCross
from tensorflow.python import SparseDenseCwiseAdd
from tensorflow.python import SparseDenseCwiseDiv
from tensorflow.python import SparseDenseCwiseMul
from tensorflow.python import SparseFillEmptyRows
from tensorflow.python import SparseFillEmptyRowsGrad
from tensorflow.python import SparseMatMul
from tensorflow.python import SparseReduceMax
from tensorflow.python import SparseReduceMaxSparse
from tensorflow.python import SparseReduceSum
from tensorflow.python import SparseReduceSumSparse
from tensorflow.python import SparseReorder
from tensorflow.python import SparseReshape
from tensorflow.python import SparseSegmentMean
from tensorflow.python import SparseSegmentMeanGrad
from tensorflow.python import SparseSegmentMeanWithNumSegments
from tensorflow.python import SparseSegmentSqrtN
from tensorflow.python import SparseSegmentSqrtNGrad
from tensorflow.python import SparseSegmentSqrtNWithNumSegments
from tensorflow.python import SparseSegmentSum
from tensorflow.python import SparseSegmentSumWithNumSegments
from tensorflow.python import SparseSlice
from tensorflow.python import SparseSliceGrad
from tensorflow.python import SparseSoftmax
from tensorflow.python import SparseSparseMaximum
from tensorflow.python import SparseSparseMinimum
from tensorflow.python import SparseSplit
from tensorflow.python import SparseTensorDenseAdd
from tensorflow.python import SparseTensorDenseMatMul
from tensorflow.python import SparseToDense
from tensorflow.python import Split
from tensorflow.python import SplitV
from tensorflow.python import Sqrt
from tensorflow.python import SqrtGrad
from tensorflow.python import Square
from tensorflow.python import SquaredDifference
from tensorflow.python import Squeeze
from tensorflow.python import Stack
from tensorflow.python import StackClose
from tensorflow.python import StackCloseV2
from tensorflow.python import StackPop
from tensorflow.python import StackPopV2
from tensorflow.python import StackPush
from tensorflow.python import StackPushV2
from tensorflow.python import StackV2
from tensorflow.python import Stage
from tensorflow.python import StageClear
from tensorflow.python import StagePeek
from tensorflow.python import StageSize
from tensorflow.python import StaticRegexFullMatch
from tensorflow.python import StaticRegexReplace
from tensorflow.python import StopGradient
from tensorflow.python import StridedSlice
from tensorflow.python import StridedSliceAssign
from tensorflow.python import StridedSliceGrad
from tensorflow.python import StringFormat
from tensorflow.python import StringJoin
from tensorflow.python import StringLength
from tensorflow.python import StringLower
from tensorflow.python import StringSplit
from tensorflow.python import StringSplitV2
from tensorflow.python import StringStrip
from tensorflow.python import StringToHashBucket
from tensorflow.python import StringToHashBucketFast
from tensorflow.python import StringToHashBucketStrong
from tensorflow.python import StringToNumber
from tensorflow.python import StringUpper
from tensorflow.python import Sub
from tensorflow.python import Substr
from tensorflow.python import Sum
from tensorflow.python import Svd
from tensorflow.python import TFRecordReaderV2
from tensorflow.python import TakeManySparseFromTensorsMap
from tensorflow.python import Tan
from tensorflow.python import Tanh
from tensorflow.python import TanhGrad
from tensorflow.python import TensorArrayClose
from tensorflow.python import TensorArrayCloseV2
from tensorflow.python import TensorArrayCloseV3
from tensorflow.python import TensorArrayConcat
from tensorflow.python import TensorArrayConcatV2
from tensorflow.python import TensorArrayConcatV3
from tensorflow.python import TensorArrayGather
from tensorflow.python import TensorArrayGatherV2
from tensorflow.python import TensorArrayGatherV3
from tensorflow.python import TensorArrayGrad
from tensorflow.python import TensorArrayGradV2
from tensorflow.python import TensorArrayGradV3
from tensorflow.python import TensorArrayGradWithShape
from tensorflow.python import TensorArrayPack
from tensorflow.python import TensorArrayRead
from tensorflow.python import TensorArrayReadV2
from tensorflow.python import TensorArrayReadV3
from tensorflow.python import TensorArrayScatter
from tensorflow.python import TensorArrayScatterV2
from tensorflow.python import TensorArrayScatterV3
from tensorflow.python import TensorArraySize
from tensorflow.python import TensorArraySizeV2
from tensorflow.python import TensorArraySizeV3
from tensorflow.python import TensorArraySplit
from tensorflow.python import TensorArraySplitV2
from tensorflow.python import TensorArraySplitV3
from tensorflow.python import TensorArrayUnpack
from tensorflow.python import TensorArrayV2
from tensorflow.python import TensorArrayV3
from tensorflow.python import TensorArrayWrite
from tensorflow.python import TensorArrayWriteV2
from tensorflow.python import TensorArrayWriteV3
from tensorflow.python import TensorScatterAdd
from tensorflow.python import TensorScatterSub
from tensorflow.python import TensorScatterUpdate
from tensorflow.python import TensorStridedSliceUpdate
from tensorflow.python import TextLineReaderV2
from tensorflow.python import Tile
from tensorflow.python import TileGrad
from tensorflow.python import Transpose
from tensorflow.python import TridiagonalMatMul
from tensorflow.python import TridiagonalSolve
from tensorflow.python import TruncateDiv
from tensorflow.python import TruncateMod
from tensorflow.python import TruncatedNormal
from tensorflow.python import Unbatch
from tensorflow.python import UnbatchGrad
from tensorflow.python import UnicodeDecode
from tensorflow.python import UnicodeDecodeWithOffsets
from tensorflow.python import UnicodeEncode
from tensorflow.python import UnicodeScript
from tensorflow.python import UnicodeTranscode
from tensorflow.python import Unique
from tensorflow.python import UniqueV2
from tensorflow.python import UniqueWithCounts
from tensorflow.python import UniqueWithCountsV2
from tensorflow.python import Unpack
from tensorflow.python import UnravelIndex
from tensorflow.python import UnsortedSegmentMax
from tensorflow.python import UnsortedSegmentMin
from tensorflow.python import UnsortedSegmentProd
from tensorflow.python import UnsortedSegmentSum
from tensorflow.python import Unstage
from tensorflow.python import UpperBound
from tensorflow.python import Where
from tensorflow.python import WholeFileReaderV2
from tensorflow.python import WriteFile
from tensorflow.python import Xdivy
from tensorflow.python import Xlogy
from tensorflow.python import ZerosLike
from tensorflow.python import Zeta
from tensorflow.python.ops.array_ops import Identity
from tensorflow.python.ops.bitwise_ops import BitwiseAnd
from tensorflow.python.ops.bitwise_ops import BitwiseOr
from tensorflow.python.ops.bitwise_ops import BitwiseXor
from tensorflow.python.ops.bitwise_ops import Invert
from tensorflow.python.ops.bitwise_ops import LeftShift
from tensorflow.python.ops.bitwise_ops import PopulationCount
from tensorflow.python.ops.bitwise_ops import RightShift
from tensorflow.python.ops.control_flow_grad import Abort
from tensorflow.python.ops.control_flow_grad import ControlTrigger
from tensorflow.python.ops.control_flow_grad import Enter
from tensorflow.python.ops.control_flow_grad import Exit
from tensorflow.python.ops.control_flow_grad import LoopCond
from tensorflow.python.ops.control_flow_grad import Merge
from tensorflow.python.ops.control_flow_grad import NextIteration
from tensorflow.python.ops.control_flow_grad import NoOp
from tensorflow.python.ops.control_flow_grad import RefEnter
from tensorflow.python.ops.control_flow_grad import RefExit
from tensorflow.python.ops.control_flow_grad import RefMerge
from tensorflow.python.ops.control_flow_grad import RefNextIteration
from tensorflow.python.ops.control_flow_grad import RefSelect
from tensorflow.python.ops.control_flow_grad import RefSwitch
from tensorflow.python.ops.control_flow_grad import Switch
from tensorflow.python.ops.data_flow_ops import TensorArray
from tensorflow.python.ops.gen_audio_ops import AudioSpectrogram
from tensorflow.python.ops.gen_audio_ops import DecodeWav
from tensorflow.python.ops.gen_audio_ops import EncodeWav
from tensorflow.python.ops.gen_audio_ops import Mfcc
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesAggregateStats
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesBucketize
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesCalculateBestFeatureSplit
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesCalculateBestGainsPerFeature
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesCenterBias
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesCreateEnsemble
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesCreateQuantileStreamResource
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesDeserializeEnsemble
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesEnsembleResourceHandleOp
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesExampleDebugOutputs
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesGetEnsembleStates
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesMakeQuantileSummaries
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesMakeStatsSummary
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesPredict
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesQuantileStreamResourceAddSummaries
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesQuantileStreamResourceDeserialize
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesQuantileStreamResourceFlush
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesQuantileStreamResourceGetBucketBoundaries
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesQuantileStreamResourceHandleOp
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesSerializeEnsemble
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesTrainingPredict
from tensorflow.python.ops.gen_boosted_trees_ops import BoostedTreesUpdateEnsemble
from tensorflow.python.ops.gen_boosted_trees_ops import IsBoostedTreesEnsembleInitialized
from tensorflow.python.ops.gen_boosted_trees_ops import IsBoostedTreesQuantileStreamResourceInitialized
from tensorflow.python.ops.gen_candidate_sampling_ops import AllCandidateSampler
from tensorflow.python.ops.gen_candidate_sampling_ops import ComputeAccidentalHits
from tensorflow.python.ops.gen_candidate_sampling_ops import FixedUnigramCandidateSampler
from tensorflow.python.ops.gen_candidate_sampling_ops import LearnedUnigramCandidateSampler
from tensorflow.python.ops.gen_candidate_sampling_ops import LogUniformCandidateSampler
from tensorflow.python.ops.gen_candidate_sampling_ops import ThreadUnsafeUnigramCandidateSampler
from tensorflow.python.ops.gen_candidate_sampling_ops import UniformCandidateSampler
from tensorflow.python.ops.gen_checkpoint_ops import GenerateVocabRemapping
from tensorflow.python.ops.gen_checkpoint_ops import LoadAndRemapMatrix
from tensorflow.python.ops.gen_collective_ops import CollectiveBcastRecv
from tensorflow.python.ops.gen_collective_ops import CollectiveBcastSend
from tensorflow.python.ops.gen_collective_ops import CollectiveGather
from tensorflow.python.ops.gen_collective_ops import CollectiveReduce
from tensorflow.python.ops.gen_ctc_ops import CTCBeamSearchDecoder
from tensorflow.python.ops.gen_ctc_ops import CTCGreedyDecoder
from tensorflow.python.ops.gen_ctc_ops import CTCLoss
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNN
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNBackprop
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNBackpropV2
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNBackpropV3
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNCanonicalToParams
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNParamsSize
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNParamsToCanonical
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNV2
from tensorflow.python.ops.gen_cudnn_rnn_ops import CudnnRNNV3
from tensorflow.python.ops.gen_data_flow_ops import Barrier
from tensorflow.python.ops.gen_data_flow_ops import ConditionalAccumulator
from tensorflow.python.ops.gen_data_flow_ops import FIFOQueue
from tensorflow.python.ops.gen_data_flow_ops import PaddingFIFOQueue
from tensorflow.python.ops.gen_data_flow_ops import PriorityQueue
from tensorflow.python.ops.gen_data_flow_ops import RandomShuffleQueue
from tensorflow.python.ops.gen_data_flow_ops import RecordInput
from tensorflow.python.ops.gen_data_flow_ops import SparseConditionalAccumulator
from tensorflow.python.ops.gen_dataset_ops import AnonymousIterator
from tensorflow.python.ops.gen_dataset_ops import AnonymousIteratorV2
from tensorflow.python.ops.gen_dataset_ops import BatchDataset
from tensorflow.python.ops.gen_dataset_ops import BatchDatasetV2
from tensorflow.python.ops.gen_dataset_ops import CacheDataset
from tensorflow.python.ops.gen_dataset_ops import ConcatenateDataset
from tensorflow.python.ops.gen_dataset_ops import DatasetToGraph
from tensorflow.python.ops.gen_dataset_ops import DatasetToSingleElement
from tensorflow.python.ops.gen_dataset_ops import DeleteIterator
from tensorflow.python.ops.gen_dataset_ops import DeserializeIterator
from tensorflow.python.ops.gen_dataset_ops import FilterByLastComponentDataset
from tensorflow.python.ops.gen_dataset_ops import FilterDataset
from tensorflow.python.ops.gen_dataset_ops import FixedLengthRecordDataset
from tensorflow.python.ops.gen_dataset_ops import FixedLengthRecordDatasetV2
from tensorflow.python.ops.gen_dataset_ops import FlatMapDataset
from tensorflow.python.ops.gen_dataset_ops import GeneratorDataset
from tensorflow.python.ops.gen_dataset_ops import InterleaveDataset
from tensorflow.python.ops.gen_dataset_ops import Iterator
from tensorflow.python.ops.gen_dataset_ops import IteratorFromStringHandle
from tensorflow.python.ops.gen_dataset_ops import IteratorFromStringHandleV2
from tensorflow.python.ops.gen_dataset_ops import IteratorGetNext
from tensorflow.python.ops.gen_dataset_ops import IteratorGetNextAsOptional
from tensorflow.python.ops.gen_dataset_ops import IteratorGetNextSync
from tensorflow.python.ops.gen_dataset_ops import IteratorToStringHandle
from tensorflow.python.ops.gen_dataset_ops import IteratorV2
from tensorflow.python.ops.gen_dataset_ops import MakeIterator
from tensorflow.python.ops.gen_dataset_ops import MapDataset
from tensorflow.python.ops.gen_dataset_ops import MapDefun
from tensorflow.python.ops.gen_dataset_ops import ModelDataset
from tensorflow.python.ops.gen_dataset_ops import MultiDeviceIterator
from tensorflow.python.ops.gen_dataset_ops import MultiDeviceIteratorFromStringHandle
from tensorflow.python.ops.gen_dataset_ops import MultiDeviceIteratorGetNextFromShard
from tensorflow.python.ops.gen_dataset_ops import MultiDeviceIteratorInit
from tensorflow.python.ops.gen_dataset_ops import MultiDeviceIteratorToStringHandle
from tensorflow.python.ops.gen_dataset_ops import OneShotIterator
from tensorflow.python.ops.gen_dataset_ops import OptimizeDataset
from tensorflow.python.ops.gen_dataset_ops import OptionalFromValue
from tensorflow.python.ops.gen_dataset_ops import OptionalGetValue
from tensorflow.python.ops.gen_dataset_ops import OptionalHasValue
from tensorflow.python.ops.gen_dataset_ops import OptionalNone
from tensorflow.python.ops.gen_dataset_ops import PaddedBatchDataset
from tensorflow.python.ops.gen_dataset_ops import PaddedBatchDatasetV2
from tensorflow.python.ops.gen_dataset_ops import ParallelInterleaveDatasetV2
from tensorflow.python.ops.gen_dataset_ops import ParallelMapDataset
from tensorflow.python.ops.gen_dataset_ops import PrefetchDataset
from tensorflow.python.ops.gen_dataset_ops import RangeDataset
from tensorflow.python.ops.gen_dataset_ops import ReduceDataset
from tensorflow.python.ops.gen_dataset_ops import RepeatDataset
from tensorflow.python.ops.gen_dataset_ops import SerializeIterator
from tensorflow.python.ops.gen_dataset_ops import ShardDataset
from tensorflow.python.ops.gen_dataset_ops import ShuffleAndRepeatDataset
from tensorflow.python.ops.gen_dataset_ops import ShuffleDataset
from tensorflow.python.ops.gen_dataset_ops import SkipDataset
from tensorflow.python.ops.gen_dataset_ops import SparseTensorSliceDataset
from tensorflow.python.ops.gen_dataset_ops import TFRecordDataset
from tensorflow.python.ops.gen_dataset_ops import TakeDataset
from tensorflow.python.ops.gen_dataset_ops import TensorDataset
from tensorflow.python.ops.gen_dataset_ops import TensorSliceDataset
from tensorflow.python.ops.gen_dataset_ops import TextLineDataset
from tensorflow.python.ops.gen_dataset_ops import UnwrapDatasetVariant
from tensorflow.python.ops.gen_dataset_ops import WindowDataset
from tensorflow.python.ops.gen_dataset_ops import WrapDatasetVariant
from tensorflow.python.ops.gen_dataset_ops import ZipDataset
from tensorflow.python.ops.gen_decode_proto_ops import DecodeProtoV2
from tensorflow.python.ops.gen_encode_proto_ops import EncodeProto
from tensorflow.python.ops.gen_experimental_dataset_ops import ChooseFastestBranchDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalAssertNextDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalAutoShardDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalBytesProducedStatsDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalCSVDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalChooseFastestDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalDatasetCardinality
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalDatasetToTFRecord
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalDenseToSparseBatchDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalDirectedInterleaveDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalGroupByReducerDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalGroupByWindowDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalIdentityIndexedDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalIgnoreErrorsDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalIndexedDatasetGet
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalIndexedDatasetMaterialize
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalIteratorGetDevice
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalLMDBDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalLatencyStatsDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalMapAndBatchDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalMapDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalMatchingFilesDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalMaterializedIndexDatasetHandle
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalMaxIntraOpParallelismDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalNonSerializableDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalNumaMapAndBatchDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalParallelInterleaveDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalParseExampleDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalPrivateThreadPoolDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalRandomDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalRebatchDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalScanDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalSetStatsAggregatorDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalSleepDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalSlidingWindowDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalSqlDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalStatsAggregatorHandle
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalStatsAggregatorSummary
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalTakeWhileDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalThreadPoolDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalThreadPoolHandle
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalUnbatchDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import ExperimentalUniqueDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import SamplingDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import SnapshotDataset
from tensorflow.python.ops.gen_experimental_dataset_ops import StatsAggregatorHandleV2
from tensorflow.python.ops.gen_experimental_dataset_ops import StatsAggregatorSetSummaryWriter
from tensorflow.python.ops.gen_functional_ops import Case
from tensorflow.python.ops.gen_functional_ops import FakeParam
from tensorflow.python.ops.gen_functional_ops import For
from tensorflow.python.ops.gen_functional_ops import If
from tensorflow.python.ops.gen_functional_ops import PartitionedCall
from tensorflow.python.ops.gen_functional_ops import RemoteCall
from tensorflow.python.ops.gen_functional_ops import StatefulPartitionedCall
from tensorflow.python.ops.gen_functional_ops import StatelessIf
from tensorflow.python.ops.gen_functional_ops import StatelessWhile
from tensorflow.python.ops.gen_functional_ops import SymbolicGradient
from tensorflow.python.ops.gen_functional_ops import While
from tensorflow.python.ops.gen_image_ops import AdjustContrast
from tensorflow.python.ops.gen_image_ops import AdjustContrastv2
from tensorflow.python.ops.gen_image_ops import AdjustHue
from tensorflow.python.ops.gen_image_ops import AdjustSaturation
from tensorflow.python.ops.gen_image_ops import CombinedNonMaxSuppression
from tensorflow.python.ops.gen_image_ops import CropAndResize
from tensorflow.python.ops.gen_image_ops import CropAndResizeGradBoxes
from tensorflow.python.ops.gen_image_ops import CropAndResizeGradImage
from tensorflow.python.ops.gen_image_ops import DecodeAndCropJpeg
from tensorflow.python.ops.gen_image_ops import DecodeBmp
from tensorflow.python.ops.gen_image_ops import DecodeGif
from tensorflow.python.ops.gen_image_ops import DecodeJpeg
from tensorflow.python.ops.gen_image_ops import DecodePng
from tensorflow.python.ops.gen_image_ops import DrawBoundingBoxes
from tensorflow.python.ops.gen_image_ops import DrawBoundingBoxesV2
from tensorflow.python.ops.gen_image_ops import EncodeJpeg
from tensorflow.python.ops.gen_image_ops import EncodeJpegVariableQuality
from tensorflow.python.ops.gen_image_ops import EncodePng
from tensorflow.python.ops.gen_image_ops import ExtractGlimpse
from tensorflow.python.ops.gen_image_ops import ExtractJpegShape
from tensorflow.python.ops.gen_image_ops import HSVToRGB
from tensorflow.python.ops.gen_image_ops import NonMaxSuppression
from tensorflow.python.ops.gen_image_ops import NonMaxSuppressionV2
from tensorflow.python.ops.gen_image_ops import NonMaxSuppressionV3
from tensorflow.python.ops.gen_image_ops import NonMaxSuppressionV4
from tensorflow.python.ops.gen_image_ops import NonMaxSuppressionWithOverlaps
from tensorflow.python.ops.gen_image_ops import QuantizedResizeBilinear
from tensorflow.python.ops.gen_image_ops import RGBToHSV
from tensorflow.python.ops.gen_image_ops import RandomCrop
from tensorflow.python.ops.gen_image_ops import ResizeArea
from tensorflow.python.ops.gen_image_ops import ResizeBicubic
from tensorflow.python.ops.gen_image_ops import ResizeBicubicGrad
from tensorflow.python.ops.gen_image_ops import ResizeBilinear
from tensorflow.python.ops.gen_image_ops import ResizeBilinearGrad
from tensorflow.python.ops.gen_image_ops import ResizeNearestNeighbor
from tensorflow.python.ops.gen_image_ops import ResizeNearestNeighborGrad
from tensorflow.python.ops.gen_image_ops import SampleDistortedBoundingBox
from tensorflow.python.ops.gen_image_ops import SampleDistortedBoundingBoxV2
from tensorflow.python.ops.gen_image_ops import ScaleAndTranslate
from tensorflow.python.ops.gen_image_ops import ScaleAndTranslateGrad
from tensorflow.python.ops.gen_io_ops import FixedLengthRecordReader
from tensorflow.python.ops.gen_io_ops import IdentityReader
from tensorflow.python.ops.gen_io_ops import LMDBReader
from tensorflow.python.ops.gen_io_ops import TFRecordReader
from tensorflow.python.ops.gen_io_ops import TextLineReader
from tensorflow.python.ops.gen_io_ops import WholeFileReader
from tensorflow.python.ops.gen_list_ops import EmptyTensorList
from tensorflow.python.ops.gen_list_ops import TensorListConcat
from tensorflow.python.ops.gen_list_ops import TensorListConcatLists
from tensorflow.python.ops.gen_list_ops import TensorListConcatV2
from tensorflow.python.ops.gen_list_ops import TensorListElementShape
from tensorflow.python.ops.gen_list_ops import TensorListFromTensor
from tensorflow.python.ops.gen_list_ops import TensorListGather
from tensorflow.python.ops.gen_list_ops import TensorListGetItem
from tensorflow.python.ops.gen_list_ops import TensorListLength
from tensorflow.python.ops.gen_list_ops import TensorListPopBack
from tensorflow.python.ops.gen_list_ops import TensorListPushBack
from tensorflow.python.ops.gen_list_ops import TensorListPushBackBatch
from tensorflow.python.ops.gen_list_ops import TensorListReserve
from tensorflow.python.ops.gen_list_ops import TensorListResize
from tensorflow.python.ops.gen_list_ops import TensorListScatter
from tensorflow.python.ops.gen_list_ops import TensorListScatterIntoExistingList
from tensorflow.python.ops.gen_list_ops import TensorListScatterV2
from tensorflow.python.ops.gen_list_ops import TensorListSetItem
from tensorflow.python.ops.gen_list_ops import TensorListSplit
from tensorflow.python.ops.gen_list_ops import TensorListStack
from tensorflow.python.ops.gen_logging_ops import Assert
from tensorflow.python.ops.gen_logging_ops import AudioSummary
from tensorflow.python.ops.gen_logging_ops import AudioSummaryV2
from tensorflow.python.ops.gen_logging_ops import HistogramSummary
from tensorflow.python.ops.gen_logging_ops import ImageSummary
from tensorflow.python.ops.gen_logging_ops import MergeSummary
from tensorflow.python.ops.gen_logging_ops import Print
from tensorflow.python.ops.gen_logging_ops import PrintV2
from tensorflow.python.ops.gen_logging_ops import ScalarSummary
from tensorflow.python.ops.gen_logging_ops import TensorSummary
from tensorflow.python.ops.gen_logging_ops import TensorSummaryV2
from tensorflow.python.ops.gen_logging_ops import Timestamp
from tensorflow.python.ops.gen_lookup_ops import HashTable
from tensorflow.python.ops.gen_lookup_ops import HashTableV2
from tensorflow.python.ops.gen_lookup_ops import InitializeTable
from tensorflow.python.ops.gen_lookup_ops import InitializeTableFromTextFile
from tensorflow.python.ops.gen_lookup_ops import InitializeTableFromTextFileV2
from tensorflow.python.ops.gen_lookup_ops import InitializeTableV2
from tensorflow.python.ops.gen_lookup_ops import LookupTableExport
from tensorflow.python.ops.gen_lookup_ops import LookupTableExportV2
from tensorflow.python.ops.gen_lookup_ops import LookupTableFind
from tensorflow.python.ops.gen_lookup_ops import LookupTableFindV2
from tensorflow.python.ops.gen_lookup_ops import LookupTableImport
from tensorflow.python.ops.gen_lookup_ops import LookupTableImportV2
from tensorflow.python.ops.gen_lookup_ops import LookupTableInsert
from tensorflow.python.ops.gen_lookup_ops import LookupTableInsertV2
from tensorflow.python.ops.gen_lookup_ops import LookupTableRemoveV2
from tensorflow.python.ops.gen_lookup_ops import LookupTableSize
from tensorflow.python.ops.gen_lookup_ops import LookupTableSizeV2
from tensorflow.python.ops.gen_lookup_ops import MutableDenseHashTable
from tensorflow.python.ops.gen_lookup_ops import MutableDenseHashTableV2
from tensorflow.python.ops.gen_lookup_ops import MutableHashTable
from tensorflow.python.ops.gen_lookup_ops import MutableHashTableOfTensors
from tensorflow.python.ops.gen_lookup_ops import MutableHashTableOfTensorsV2
from tensorflow.python.ops.gen_lookup_ops import MutableHashTableV2
from tensorflow.python.ops.gen_manip_ops import Roll
from tensorflow.python.ops.gen_nccl_ops import NcclAllReduce
from tensorflow.python.ops.gen_nccl_ops import NcclBroadcast
from tensorflow.python.ops.gen_nccl_ops import NcclReduce
from tensorflow.python.ops.gen_nn_ops import AvgPool
from tensorflow.python.ops.gen_nn_ops import AvgPool3D
from tensorflow.python.ops.gen_nn_ops import AvgPool3DGrad
from tensorflow.python.ops.gen_nn_ops import AvgPoolGrad
from tensorflow.python.ops.gen_nn_ops import BatchNormWithGlobalNormalization
from tensorflow.python.ops.gen_nn_ops import BatchNormWithGlobalNormalizationGrad
from tensorflow.python.ops.gen_nn_ops import BiasAdd
from tensorflow.python.ops.gen_nn_ops import BiasAddGrad
from tensorflow.python.ops.gen_nn_ops import BiasAddV1
from tensorflow.python.ops.gen_nn_ops import Conv2D
from tensorflow.python.ops.gen_nn_ops import Conv2DBackpropFilter
from tensorflow.python.ops.gen_nn_ops import Conv2DBackpropInput
from tensorflow.python.ops.gen_nn_ops import Conv3D
from tensorflow.python.ops.gen_nn_ops import Conv3DBackpropFilter
from tensorflow.python.ops.gen_nn_ops import Conv3DBackpropFilterV2
from tensorflow.python.ops.gen_nn_ops import Conv3DBackpropInput
from tensorflow.python.ops.gen_nn_ops import Conv3DBackpropInputV2
from tensorflow.python.ops.gen_nn_ops import DataFormatDimMap
from tensorflow.python.ops.gen_nn_ops import DataFormatVecPermute
from tensorflow.python.ops.gen_nn_ops import DepthwiseConv2dNative
from tensorflow.python.ops.gen_nn_ops import DepthwiseConv2dNativeBackpropFilter
from tensorflow.python.ops.gen_nn_ops import DepthwiseConv2dNativeBackpropInput
from tensorflow.python.ops.gen_nn_ops import Dilation2D
from tensorflow.python.ops.gen_nn_ops import Dilation2DBackpropFilter
from tensorflow.python.ops.gen_nn_ops import Dilation2DBackpropInput
from tensorflow.python.ops.gen_nn_ops import Elu
from tensorflow.python.ops.gen_nn_ops import EluGrad
from tensorflow.python.ops.gen_nn_ops import FractionalAvgPool
from tensorflow.python.ops.gen_nn_ops import FractionalAvgPoolGrad
from tensorflow.python.ops.gen_nn_ops import FractionalMaxPool
from tensorflow.python.ops.gen_nn_ops import FractionalMaxPoolGrad
from tensorflow.python.ops.gen_nn_ops import FusedBatchNorm
from tensorflow.python.ops.gen_nn_ops import FusedBatchNormGrad
from tensorflow.python.ops.gen_nn_ops import FusedBatchNormGradV2
from tensorflow.python.ops.gen_nn_ops import FusedBatchNormGradV3
from tensorflow.python.ops.gen_nn_ops import FusedBatchNormV2
from tensorflow.python.ops.gen_nn_ops import FusedBatchNormV3
from tensorflow.python.ops.gen_nn_ops import FusedPadConv2D
from tensorflow.python.ops.gen_nn_ops import FusedResizeAndPadConv2D
from tensorflow.python.ops.gen_nn_ops import InTopK
from tensorflow.python.ops.gen_nn_ops import InTopKV2
from tensorflow.python.ops.gen_nn_ops import L2Loss
from tensorflow.python.ops.gen_nn_ops import LRN
from tensorflow.python.ops.gen_nn_ops import LRNGrad
from tensorflow.python.ops.gen_nn_ops import LeakyRelu
from tensorflow.python.ops.gen_nn_ops import LeakyReluGrad
from tensorflow.python.ops.gen_nn_ops import LogSoftmax
from tensorflow.python.ops.gen_nn_ops import MaxPool
from tensorflow.python.ops.gen_nn_ops import MaxPool3D
from tensorflow.python.ops.gen_nn_ops import MaxPool3DGrad
from tensorflow.python.ops.gen_nn_ops import MaxPool3DGradGrad
from tensorflow.python.ops.gen_nn_ops import MaxPoolGrad
from tensorflow.python.ops.gen_nn_ops import MaxPoolGradGrad
from tensorflow.python.ops.gen_nn_ops import MaxPoolGradGradV2
from tensorflow.python.ops.gen_nn_ops import MaxPoolGradGradWithArgmax
from tensorflow.python.ops.gen_nn_ops import MaxPoolGradV2
from tensorflow.python.ops.gen_nn_ops import MaxPoolGradWithArgmax
from tensorflow.python.ops.gen_nn_ops import MaxPoolV2
from tensorflow.python.ops.gen_nn_ops import MaxPoolWithArgmax
from tensorflow.python.ops.gen_nn_ops import NthElement
from tensorflow.python.ops.gen_nn_ops import QuantizedAvgPool
from tensorflow.python.ops.gen_nn_ops import QuantizedBatchNormWithGlobalNormalization
from tensorflow.python.ops.gen_nn_ops import QuantizedBiasAdd
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2D
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DAndRelu
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DAndReluAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DPerChannel
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBias
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBiasAndRelu
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBiasAndReluAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBiasAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBiasSignedSumAndReluAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBiasSumAndRelu
from tensorflow.python.ops.gen_nn_ops import QuantizedConv2DWithBiasSumAndReluAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedDepthwiseConv2D
from tensorflow.python.ops.gen_nn_ops import QuantizedDepthwiseConv2DWithBias
from tensorflow.python.ops.gen_nn_ops import QuantizedDepthwiseConv2DWithBiasAndRelu
from tensorflow.python.ops.gen_nn_ops import QuantizedDepthwiseConv2DWithBiasAndReluAndRequantize
from tensorflow.python.ops.gen_nn_ops import QuantizedMaxPool
from tensorflow.python.ops.gen_nn_ops import QuantizedRelu
from tensorflow.python.ops.gen_nn_ops import QuantizedRelu6
from tensorflow.python.ops.gen_nn_ops import QuantizedReluX
from tensorflow.python.ops.gen_nn_ops import Relu
from tensorflow.python.ops.gen_nn_ops import Relu6
from tensorflow.python.ops.gen_nn_ops import Relu6Grad
from tensorflow.python.ops.gen_nn_ops import ReluGrad
from tensorflow.python.ops.gen_nn_ops import Selu
from tensorflow.python.ops.gen_nn_ops import SeluGrad
from tensorflow.python.ops.gen_nn_ops import Softmax
from tensorflow.python.ops.gen_nn_ops import SoftmaxCrossEntropyWithLogits
from tensorflow.python.ops.gen_nn_ops import Softplus
from tensorflow.python.ops.gen_nn_ops import SoftplusGrad
from tensorflow.python.ops.gen_nn_ops import Softsign
from tensorflow.python.ops.gen_nn_ops import SoftsignGrad
from tensorflow.python.ops.gen_nn_ops import SparseSoftmaxCrossEntropyWithLogits
from tensorflow.python.ops.gen_nn_ops import TopK
from tensorflow.python.ops.gen_nn_ops import TopKV2
from tensorflow.python.ops.gen_ragged_array_ops import RaggedGather
from tensorflow.python.ops.gen_ragged_conversion_ops import RaggedTensorFromVariant
from tensorflow.python.ops.gen_ragged_conversion_ops import RaggedTensorToSparse
from tensorflow.python.ops.gen_ragged_conversion_ops import RaggedTensorToVariant
from tensorflow.python.ops.gen_ragged_math_ops import RaggedRange
from tensorflow.python.ops.gen_resource_variable_ops import AssignAddVariableOp
from tensorflow.python.ops.gen_resource_variable_ops import AssignSubVariableOp
from tensorflow.python.ops.gen_resource_variable_ops import AssignVariableOp
from tensorflow.python.ops.gen_resource_variable_ops import ConsumeMutexLock
from tensorflow.python.ops.gen_resource_variable_ops import DestroyResourceOp
from tensorflow.python.ops.gen_resource_variable_ops import MutexLock
from tensorflow.python.ops.gen_resource_variable_ops import MutexV2
from tensorflow.python.ops.gen_resource_variable_ops import ReadVariableOp
from tensorflow.python.ops.gen_resource_variable_ops import ResourceGather
from tensorflow.python.ops.gen_resource_variable_ops import ResourceGatherNd
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterAdd
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterDiv
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterMax
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterMin
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterMul
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterSub
from tensorflow.python.ops.gen_resource_variable_ops import ResourceScatterUpdate
from tensorflow.python.ops.gen_resource_variable_ops import VarHandleOp
from tensorflow.python.ops.gen_resource_variable_ops import VarIsInitializedOp
from tensorflow.python.ops.gen_resource_variable_ops import VariableShape
from tensorflow.python.ops.gen_script_ops import EagerPyFunc
from tensorflow.python.ops.gen_script_ops import PyFunc
from tensorflow.python.ops.gen_script_ops import PyFuncStateless
from tensorflow.python.ops.gen_sdca_ops import SdcaFprint
from tensorflow.python.ops.gen_sdca_ops import SdcaOptimizer
from tensorflow.python.ops.gen_sdca_ops import SdcaOptimizerV2
from tensorflow.python.ops.gen_sdca_ops import SdcaShrinkL1
from tensorflow.python.ops.gen_set_ops import DenseToDenseSetOperation
from tensorflow.python.ops.gen_set_ops import DenseToSparseSetOperation
from tensorflow.python.ops.gen_set_ops import SetSize
from tensorflow.python.ops.gen_set_ops import SparseToSparseSetOperation
from tensorflow.python.ops.gen_spectral_ops import BatchFFT
from tensorflow.python.ops.gen_spectral_ops import BatchFFT2D
from tensorflow.python.ops.gen_spectral_ops import BatchFFT3D
from tensorflow.python.ops.gen_spectral_ops import BatchIFFT
from tensorflow.python.ops.gen_spectral_ops import BatchIFFT2D
from tensorflow.python.ops.gen_spectral_ops import BatchIFFT3D
from tensorflow.python.ops.gen_spectral_ops import FFT
from tensorflow.python.ops.gen_spectral_ops import FFT2D
from tensorflow.python.ops.gen_spectral_ops import FFT3D
from tensorflow.python.ops.gen_spectral_ops import IFFT
from tensorflow.python.ops.gen_spectral_ops import IFFT2D
from tensorflow.python.ops.gen_spectral_ops import IFFT3D
from tensorflow.python.ops.gen_spectral_ops import IRFFT
from tensorflow.python.ops.gen_spectral_ops import IRFFT2D
from tensorflow.python.ops.gen_spectral_ops import IRFFT3D
from tensorflow.python.ops.gen_spectral_ops import RFFT
from tensorflow.python.ops.gen_spectral_ops import RFFT2D
from tensorflow.python.ops.gen_spectral_ops import RFFT3D
from tensorflow.python.ops.gen_state_ops import Assign
from tensorflow.python.ops.gen_state_ops import AssignAdd
from tensorflow.python.ops.gen_state_ops import AssignSub
from tensorflow.python.ops.gen_state_ops import CountUpTo
from tensorflow.python.ops.gen_state_ops import DestroyTemporaryVariable
from tensorflow.python.ops.gen_state_ops import IsVariableInitialized
from tensorflow.python.ops.gen_state_ops import ResourceCountUpTo
from tensorflow.python.ops.gen_state_ops import ResourceScatterNdAdd
from tensorflow.python.ops.gen_state_ops import ResourceScatterNdSub
from tensorflow.python.ops.gen_state_ops import ResourceScatterNdUpdate
from tensorflow.python.ops.gen_state_ops import ScatterAdd
from tensorflow.python.ops.gen_state_ops import ScatterDiv
from tensorflow.python.ops.gen_state_ops import ScatterMax
from tensorflow.python.ops.gen_state_ops import ScatterMin
from tensorflow.python.ops.gen_state_ops import ScatterMul
from tensorflow.python.ops.gen_state_ops import ScatterNdAdd
from tensorflow.python.ops.gen_state_ops import ScatterNdSub
from tensorflow.python.ops.gen_state_ops import ScatterNdUpdate
from tensorflow.python.ops.gen_state_ops import ScatterSub
from tensorflow.python.ops.gen_state_ops import ScatterUpdate
from tensorflow.python.ops.gen_state_ops import TemporaryVariable
from tensorflow.python.ops.gen_state_ops import Variable
from tensorflow.python.ops.gen_state_ops import VariableV2
from tensorflow.python.ops.gen_stateful_random_ops import NonDeterministicInts
from tensorflow.python.ops.gen_stateful_random_ops import RngSkip
from tensorflow.python.ops.gen_stateful_random_ops import StatefulRandomBinomial
from tensorflow.python.ops.gen_stateful_random_ops import StatefulStandardNormal
from tensorflow.python.ops.gen_stateful_random_ops import StatefulStandardNormalV2
from tensorflow.python.ops.gen_stateful_random_ops import StatefulTruncatedNormal
from tensorflow.python.ops.gen_stateful_random_ops import StatefulUniform
from tensorflow.python.ops.gen_stateful_random_ops import StatefulUniformFullInt
from tensorflow.python.ops.gen_stateful_random_ops import StatefulUniformInt
from tensorflow.python.ops.gen_stateless_random_ops import StatelessMultinomial
from tensorflow.python.ops.gen_stateless_random_ops import StatelessRandomNormal
from tensorflow.python.ops.gen_stateless_random_ops import StatelessRandomUniform
from tensorflow.python.ops.gen_stateless_random_ops import StatelessRandomUniformInt
from tensorflow.python.ops.gen_stateless_random_ops import StatelessTruncatedNormal
from tensorflow.python.ops.gen_summary_ops import CloseSummaryWriter
from tensorflow.python.ops.gen_summary_ops import CreateSummaryDbWriter
from tensorflow.python.ops.gen_summary_ops import CreateSummaryFileWriter
from tensorflow.python.ops.gen_summary_ops import FlushSummaryWriter
from tensorflow.python.ops.gen_summary_ops import ImportEvent
from tensorflow.python.ops.gen_summary_ops import SummaryWriter
from tensorflow.python.ops.gen_summary_ops import WriteAudioSummary
from tensorflow.python.ops.gen_summary_ops import WriteGraphSummary
from tensorflow.python.ops.gen_summary_ops import WriteHistogramSummary
from tensorflow.python.ops.gen_summary_ops import WriteImageSummary
from tensorflow.python.ops.gen_summary_ops import WriteRawProtoSummary
from tensorflow.python.ops.gen_summary_ops import WriteScalarSummary
from tensorflow.python.ops.gen_summary_ops import WriteSummary
from tensorflow.python.ops.gen_tpu_ops import AllToAll
from tensorflow.python.ops.gen_tpu_ops import CollectivePermute
from tensorflow.python.ops.gen_tpu_ops import ConfigureDistributedTPU
from tensorflow.python.ops.gen_tpu_ops import CrossReplicaSum
from tensorflow.python.ops.gen_tpu_ops import EnqueueTPUEmbeddingIntegerBatch
from tensorflow.python.ops.gen_tpu_ops import EnqueueTPUEmbeddingSparseBatch
from tensorflow.python.ops.gen_tpu_ops import EnqueueTPUEmbeddingSparseTensorBatch
from tensorflow.python.ops.gen_tpu_ops import InfeedDequeue
from tensorflow.python.ops.gen_tpu_ops import InfeedDequeueTuple
from tensorflow.python.ops.gen_tpu_ops import InfeedEnqueue
from tensorflow.python.ops.gen_tpu_ops import InfeedEnqueuePrelinearizedBuffer
from tensorflow.python.ops.gen_tpu_ops import InfeedEnqueueTuple
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingADAMParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingADAMParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingAdadeltaParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingAdadeltaParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingAdagradParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingAdagradParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingCenteredRMSPropParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingFTRLParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingFTRLParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingMDLAdagradLightParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingMomentumParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingMomentumParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingProximalAdagradParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingProximalAdagradParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingRMSPropParameters
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingRMSPropParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import LoadTPUEmbeddingStochasticGradientDescentParameters
from tensorflow.python.ops.gen_tpu_ops import OutfeedDequeue
from tensorflow.python.ops.gen_tpu_ops import OutfeedDequeueTuple
from tensorflow.python.ops.gen_tpu_ops import OutfeedEnqueue
from tensorflow.python.ops.gen_tpu_ops import OutfeedEnqueueTuple
from tensorflow.python.ops.gen_tpu_ops import Prelinearize
from tensorflow.python.ops.gen_tpu_ops import PrelinearizeTuple
from tensorflow.python.ops.gen_tpu_ops import RecvTPUEmbeddingActivations
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingADAMParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingADAMParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingAdadeltaParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingAdadeltaParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingAdagradParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingAdagradParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingCenteredRMSPropParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingFTRLParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingFTRLParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingMDLAdagradLightParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingMomentumParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingMomentumParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingProximalAdagradParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingProximalAdagradParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingRMSPropParameters
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingRMSPropParametersGradAccumDebug
from tensorflow.python.ops.gen_tpu_ops import RetrieveTPUEmbeddingStochasticGradientDescentParameters
from tensorflow.python.ops.gen_tpu_ops import SendTPUEmbeddingGradients
from tensorflow.python.ops.gen_tpu_ops import ShutdownDistributedTPU
from tensorflow.python.ops.gen_tpu_ops import TPUCompilationResult
from tensorflow.python.ops.gen_tpu_ops import TPUEmbeddingActivations
from tensorflow.python.ops.gen_tpu_ops import TPUOrdinalSelector
from tensorflow.python.ops.gen_tpu_ops import TPUPartitionedCall
from tensorflow.python.ops.gen_tpu_ops import TPUReplicateMetadata
from tensorflow.python.ops.gen_tpu_ops import TPUReplicatedInput
from tensorflow.python.ops.gen_tpu_ops import TPUReplicatedOutput
from tensorflow.python.ops.gen_tpu_ops import WorkerHeartbeat
from tensorflow.python.ops.gen_user_ops import Fact
from tensorflow.python.training.gen_training_ops import ApplyAdaMax
from tensorflow.python.training.gen_training_ops import ApplyAdadelta
from tensorflow.python.training.gen_training_ops import ApplyAdagrad
from tensorflow.python.training.gen_training_ops import ApplyAdagradDA
from tensorflow.python.training.gen_training_ops import ApplyAdam
from tensorflow.python.training.gen_training_ops import ApplyAddSign
from tensorflow.python.training.gen_training_ops import ApplyCenteredRMSProp
from tensorflow.python.training.gen_training_ops import ApplyFtrl
from tensorflow.python.training.gen_training_ops import ApplyFtrlV2
from tensorflow.python.training.gen_training_ops import ApplyGradientDescent
from tensorflow.python.training.gen_training_ops import ApplyMomentum
from tensorflow.python.training.gen_training_ops import ApplyPowerSign
from tensorflow.python.training.gen_training_ops import ApplyProximalAdagrad
from tensorflow.python.training.gen_training_ops import ApplyProximalGradientDescent
from tensorflow.python.training.gen_training_ops import ApplyRMSProp
from tensorflow.python.training.gen_training_ops import ResourceApplyAdaMax
from tensorflow.python.training.gen_training_ops import ResourceApplyAdadelta
from tensorflow.python.training.gen_training_ops import ResourceApplyAdagrad
from tensorflow.python.training.gen_training_ops import ResourceApplyAdagradDA
from tensorflow.python.training.gen_training_ops import ResourceApplyAdam
from tensorflow.python.training.gen_training_ops import ResourceApplyAdamWithAmsgrad
from tensorflow.python.training.gen_training_ops import ResourceApplyAddSign
from tensorflow.python.training.gen_training_ops import ResourceApplyCenteredRMSProp
from tensorflow.python.training.gen_training_ops import ResourceApplyFtrl
from tensorflow.python.training.gen_training_ops import ResourceApplyFtrlV2
from tensorflow.python.training.gen_training_ops import ResourceApplyGradientDescent
from tensorflow.python.training.gen_training_ops import ResourceApplyKerasMomentum
from tensorflow.python.training.gen_training_ops import ResourceApplyMomentum
from tensorflow.python.training.gen_training_ops import ResourceApplyPowerSign
from tensorflow.python.training.gen_training_ops import ResourceApplyProximalAdagrad
from tensorflow.python.training.gen_training_ops import ResourceApplyProximalGradientDescent
from tensorflow.python.training.gen_training_ops import ResourceApplyRMSProp
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyAdadelta
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyAdagrad
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyAdagradDA
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyCenteredRMSProp
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyFtrl
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyFtrlV2
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyKerasMomentum
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyMomentum
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyProximalAdagrad
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyProximalGradientDescent
from tensorflow.python.training.gen_training_ops import ResourceSparseApplyRMSProp
from tensorflow.python.training.gen_training_ops import SparseApplyAdadelta
from tensorflow.python.training.gen_training_ops import SparseApplyAdagrad
from tensorflow.python.training.gen_training_ops import SparseApplyAdagradDA
from tensorflow.python.training.gen_training_ops import SparseApplyCenteredRMSProp
from tensorflow.python.training.gen_training_ops import SparseApplyFtrl
from tensorflow.python.training.gen_training_ops import SparseApplyFtrlV2
from tensorflow.python.training.gen_training_ops import SparseApplyMomentum
from tensorflow.python.training.gen_training_ops import SparseApplyProximalAdagrad
from tensorflow.python.training.gen_training_ops import SparseApplyProximalGradientDescent
from tensorflow.python.training.gen_training_ops import SparseApplyRMSProp

del _print_function

import sys as _sys
from tensorflow.python.util import deprecation_wrapper as _deprecation_wrapper

if not isinstance(_sys.modules[__name__], _deprecation_wrapper.DeprecationWrapper):
  _sys.modules[__name__] = _deprecation_wrapper.DeprecationWrapper(
      _sys.modules[__name__], "raw_ops")



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Core conversion logic, serves as main point of access."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools
import imp
import sys
import types
import unittest
import weakref

import gast

from tensorflow.python.autograph import operators
from tensorflow.python.autograph import utils
from tensorflow.python.autograph.converters import arg_defaults
from tensorflow.python.autograph.converters import asserts
from tensorflow.python.autograph.converters import break_statements
from tensorflow.python.autograph.converters import call_trees
from tensorflow.python.autograph.converters import conditional_expressions
from tensorflow.python.autograph.converters import continue_statements
from tensorflow.python.autograph.converters import control_flow
from tensorflow.python.autograph.converters import directives
from tensorflow.python.autograph.converters import function_scopes
from tensorflow.python.autograph.converters import lists
from tensorflow.python.autograph.converters import logical_expressions
from tensorflow.python.autograph.converters import return_statements
from tensorflow.python.autograph.converters import side_effect_guards
from tensorflow.python.autograph.converters import slices
from tensorflow.python.autograph.core import config
from tensorflow.python.autograph.core import converter
from tensorflow.python.autograph.core import function_wrapping
from tensorflow.python.autograph.core import naming
from tensorflow.python.autograph.core import unsupported_features_checker
from tensorflow.python.autograph.lang import special_functions
from tensorflow.python.autograph.pyct import ast_util
from tensorflow.python.autograph.pyct import compiler
from tensorflow.python.autograph.pyct import inspect_utils
from tensorflow.python.autograph.pyct import origin_info
from tensorflow.python.autograph.pyct import parser
from tensorflow.python.autograph.pyct import pretty_printer
from tensorflow.python.autograph.pyct import qual_names
from tensorflow.python.autograph.pyct import templates
from tensorflow.python.autograph.pyct import transformer
from tensorflow.python.autograph.utils import ag_logging as logging
from tensorflow.python.util import tf_inspect


class _ConvertedEntityFactoryInfo(
    collections.namedtuple(
        '_ConvertedEntityFactoryInfo',
        ('module_name', 'converted_name', 'factory_factory_name', 'source_map'))
):
  """Holds metadata about a converted entity stored as a dynamic factory.

  The dynamic factory is assumed to be created by _wrap_into_dynamic_factory,
  be named `factory_factory_name` and located inside the module named as
  `module_name`.

  Attributes:
    module_name: Text, the name of the module containing the entity.
    converted_name: Text, the name of the converted entity.
    factory_factory_name: Text, the name of the dynamic factory.
    source_map: Dict.
  """

  def get_module(self):
    return sys.modules[self.module_name]

  def get_factory(self):
    assert self.module_name in sys.modules
    factory_factory = getattr(sys.modules[self.module_name],
                              self.factory_factory_name)
    return factory_factory()


# TODO(mdan): Add a garbage collection hook for cleaning up modules.
class _ConversionCache(object):
  """A hierarchical cache that uses the converted entity as weak key.

  The keys soft references (i.e. they are discarded when the key is
  destroyed). The subkeys are normal hashable values.

  This class is generic - see the call site for how the keys and values are
  defined.
  """

  def __init__(self):
    self._cache = weakref.WeakKeyDictionary()

  def has(self, key, subkey):
    if key not in self._cache:
      return False
    return subkey in self._cache[key]

  def __getitem__(self, key):
    if key not in self._cache:
      # The bucket needs to be initialized to support this usage:
      #   cache[key][subkey] = value
      self._cache[key] = {}
    return self._cache[key]


_CACHE = _ConversionCache()


# Note: strictly speaking, a simple factory might have been sufficient for
# functions. But the double factory approach allows us to control the closure
# and globals of the converted code in a cleaner fashion.
# TODO(mdan): A simple factory may be sufficient.
def _wrap_into_dynamic_factory(nodes, entity_name, factory_factory_name,
                               factory_name, closure_vars, future_features):
  """Wraps an AST into the body of a dynamic factory.

  This uses the dynamic factory (factory of factory) pattern to achieve the
  following:

   1. The inner factory, dynamically creates the entity represented by nodes.
   2. The entity is parametrized by `ag__`, the internal AutoGraph module.
   3. The outer factory creates the inner factory with a lexical scope
      in which `closure_vars` are bound local variables. This in turn allows the
      caller to control the exact closure (i.e. non-global free variables) for
      the inner factory.

  The AST is expected to define some symbol named by `entity_name`.

  Args:
    nodes: ast.AST
    entity_name: Union[Text, ast.AST]
    factory_factory_name: Text
    factory_name: Text
    closure_vars: Iterable[Text]
    future_features: Iterable[Text], see EntityInfo.future_features.

  Returns:
    ast.AST
  """
  if not isinstance(nodes, (list, tuple)):
    nodes = (nodes,)

  dummy_closure_defs = []
  for var_name in closure_vars:
    template = """
      var_name = None
    """
    dummy_closure_defs.extend(templates.replace(template, var_name=var_name))

  if future_features:
    future_imports = gast.ImportFrom(
        module='__future__',
        names=[gast.alias(name=name, asname=None) for name in future_features],
        level=0)
  else:
    future_imports = []

  # These dummy symbol declarations create local fariables in a function scope,
  # so that the Python parser correctly marks them as free non-global variables
  # upon load (that is, it creates cell slots for each symbol). Their values are
  # not used, as the cells are swapped with the original entity's cells after
  # the code has been loaded.
  template = """
    future_imports
    def factory_factory_name():
      dummy_closure_defs
      def factory_name(ag__, ag_source_map__, ag_module__):
        entity_defs
        entity_name.ag_source_map = ag_source_map__
        entity_name.ag_module = ag_module__
        entity_name.autograph_info__ = {}
        return entity_name
      return factory_name
  """
  return templates.replace(
      template,
      future_imports=future_imports,
      factory_factory_name=factory_factory_name,
      factory_name=factory_name,
      dummy_closure_defs=dummy_closure_defs,
      entity_defs=nodes,
      entity_name=entity_name)


def _convert_with_cache(entity, program_ctx, free_nonglobal_var_names):
  """Returns a (possibly cached) factory for the converted result of entity."""
  # The cache key is the entity's code object if it defined one, otherwise it's
  # the entity itself. Keying by the code object allows caching of functions
  # that are dynamically created e.g. in a loop.
  if hasattr(entity, '__code__'):
    key = entity.__code__
  else:
    key = entity

  # The cache subkey encompases any conversion options on which the generated
  # code may depend.
  # The cached factory includes the necessary definitions to distinguish
  # between the global and non-global free variables. For this reason, the
  # cache subkey includes the names of the free non-globals.
  subkey = (program_ctx.options, frozenset(free_nonglobal_var_names))

  # The cache values are _ConvertedEntityFactoryInfo objects.
  if _CACHE.has(key, subkey):
    # TODO(mdan): Check whether the module is still loaded.
    converted_entity_info = _CACHE[key][subkey]
    logging.log(3, 'Cache hit for entity %s key %s subkey %s: %s', entity, key,
                subkey, converted_entity_info)
    return converted_entity_info

  logging.log(1, 'Entity %s is not cached for key %s subkey %s', entity, key,
              subkey)

  nodes, converted_name, entity_info = convert_entity_to_ast(
      entity, program_ctx)

  namer = naming.Namer(entity_info.namespace)
  factory_factory_name = namer.new_symbol('create_converted_entity_factory', ())
  factory_name = namer.new_symbol('create_converted_entity', ())
  nodes = _wrap_into_dynamic_factory(
      nodes, converted_name, factory_factory_name, factory_name,
      free_nonglobal_var_names, entity_info.future_features)

  module, _, source_map = compiler.ast_to_object(nodes, include_source_map=True)
  module_name = module.__name__

  converted_entity_info = _ConvertedEntityFactoryInfo(
      module_name=module_name,
      converted_name=converted_name,
      factory_factory_name=factory_factory_name,
      source_map=source_map)
  _CACHE[key][subkey] = converted_entity_info
  return converted_entity_info


def _instantiate(entity, converted_entity_info, free_nonglobal_var_names):
  """Creates a converted instance and binds it to match original entity."""
  factory = converted_entity_info.get_factory()

  # `factory` is currently bound to the empty module it was loaded from.
  # It must instead be bound to the globals and closure from the original
  # entity.
  if tf_inspect.isfunction(entity) or tf_inspect.ismethod(entity):
    entity_globals = entity.__globals__
    entity_closure = entity.__closure__ or ()
  elif hasattr(entity, '__module__'):
    entity_globals = sys.modules[entity.__module__].__dict__
    entity_closure = ()
  assert len(entity_closure) == len(free_nonglobal_var_names)

  # Fit the original entity's cells to match the order of factory's cells.
  original_names_and_cells = dict(zip(free_nonglobal_var_names, entity_closure))
  new_factory_cells = tuple(
      original_names_and_cells[name] for name in factory.__code__.co_freevars)

  bound_factory = types.FunctionType(
      code=factory.__code__,
      globals=entity_globals,
      name=factory.__name__,
      argdefs=(),
      closure=new_factory_cells)

  # Two other free vars: the internal "ag__" module and the source
  # map. These are wired via the parameters of the factory.
  converted_entity = bound_factory(  # pylint:disable=not-callable
      ag_internal, converted_entity_info.source_map,
      converted_entity_info.get_module())

  if tf_inspect.isfunction(entity) or tf_inspect.ismethod(entity):
    # Attach the default argument to the converted function.
    converted_entity.__defaults__ = entity.__defaults__

  return converted_entity


def convert(entity, program_ctx):
  """Converts an entity into an equivalent entity."""

  if tf_inspect.isfunction(entity) or tf_inspect.ismethod(entity):
    free_nonglobal_var_names = entity.__code__.co_freevars
  else:
    free_nonglobal_var_names = ()

  for i, name in enumerate(free_nonglobal_var_names):
    if (name == 'ag__' and
        entity.__closure__[i].cell_contents is not ag_internal):
      raise ValueError('entity {} uses the reserved symbol "{}"'.format(
          entity, name))
    # TODO(mdan): In extreme cases, other ag__ symbols may also be clobbered.

  converted_entity_info = _convert_with_cache(
      entity, program_ctx, free_nonglobal_var_names)

  return _instantiate(entity, converted_entity_info, free_nonglobal_var_names)


def is_whitelisted_for_graph(o, check_call_override=True):
  """Checks whether an entity is whitelisted for use in graph mode.

  Examples of whitelisted entities include all members of the tensorflow
  package.

  Args:
    o: A Python entity.
    check_call_override: Reserved for internal use. When set to `False`, it
      disables the rule according to which classes are whitelisted if their
      __call__ method is whitelisted.

  Returns:
    Boolean
  """
  # TODO(b/120224672): Fix this.
  if isinstance(o, functools.partial):
    # tf_inspect.getmodule(functools.partial(...)) otherwise returns None since
    # functools.partial objects do not have a __module__ attribute.
    m = functools
  else:
    m = tf_inspect.getmodule(o)

  if hasattr(m, '__name__'):
    # Builtins typically have unnamed modules.
    for prefix, in config.DEFAULT_UNCOMPILED_MODULES:
      if m.__name__.startswith(prefix + '.') or m.__name__ == prefix:
        logging.log(2, 'Whitelisted: %s: name starts with "%s"', o, prefix)
        return True

  if hasattr(o, 'autograph_info__') or hasattr(o, '__ag_compiled'):
    logging.log(2, 'Whitelisted: %s: already converted', o)
    return True

  if tf_inspect.isgeneratorfunction(o):
    logging.warn(
        'Entity {} appears to be a generator function. It will not be converted'
        ' by AutoGraph.'.format(o), 1)
    logging.log(2, 'Whitelisted: %s: generator functions are not converted', o)
    return True

  if check_call_override and hasattr(o, '__call__'):
    # Callable objects: whitelisted if their __call__ method is.
    # The type check avoids infinite recursion around the __call__ method
    # of function objects.
    if (type(o) != type(o.__call__)) and is_whitelisted_for_graph(o.__call__):  # pylint: disable=unidiomatic-typecheck
      logging.log(2, 'Whitelisted: %s: object __call__ whitelisted', o)
      return True

  owner_class = None
  if tf_inspect.ismethod(o):
    # Methods of whitelisted classes are also whitelisted, even if they are
    # bound via user subclasses.
    #
    # For example, suppose `tf.Foo` has a method called `bar`, and `baz` is
    # defined as below. `tf.Foo` is whitelisted. Then `baz.bar` is also
    # whitelisted.
    #
    #   class Custom(tf.Foo):
    #     pass
    #
    #   baz = Custom()
    #
    # For the example above, if `Custom` did overload `bar`, then it would no
    # longer be whitelisted.

    owner_class = inspect_utils.getmethodclass(o)
    if owner_class is not None:
      if issubclass(owner_class, unittest.TestCase):
        logging.log(2, 'Whitelisted: %s: method of TestCase subclass', o)
        return True

      owner_class = inspect_utils.getdefiningclass(o, owner_class)
      is_call_override = (o.__name__ == '__call__')
      if is_whitelisted_for_graph(
          owner_class, check_call_override=not is_call_override):
        logging.log(2, 'Whitelisted: %s: owner is whitelisted %s', o,
                    owner_class)
        return True

  if inspect_utils.isnamedtuple(o):
    # Due to the way they're constructed, namedtuple types cannot be converted
    # because they don't expose source code. But we assume they are safe for
    # graph mode since they are just containers.
    if tf_inspect.isclass(o) and len(o.__bases__) > 1:
      logging.warn(
          'Entity {} looks like a namedtuple subclass. Its constructor will'
          ' not be converted by AutoGraph, but if it has any custom methods,'
          ' those will be.'.format(o), 1)
    logging.log(2, 'Whitelisted: %s: named tuple', o)
    return True

  logging.log(2, 'Not whitelisted: %s: default rule', o)
  return False


# TODO(mdan): Rename to convert_*_node to avoid confusion with convert.
def convert_entity_to_ast(o, program_ctx):
  """Compile a Python entity into equivalent TensorFlow.

  Args:
    o: A Python entity.
    program_ctx: A ProgramContext object.

  Returns:
    A tuple (ast, new_name, namespace):
        * ast: An AST representing an entity with interface equivalent to `o`,
            but which when executed it creates TF a graph.
        * new_name: The symbol name under which the new entity can be found.
        * namespace: A dict mapping all symbols visible to the converted entity,
            keyed by their symbol name.

  Raises:
    ValueError: if the entity type is not supported.
  """
  logging.log(1, 'Converting %s', o)

  if tf_inspect.isclass(o):
    nodes, name, entity_info = convert_class_to_ast(o, program_ctx)
  elif tf_inspect.isfunction(o):
    nodes, name, entity_info = convert_func_to_ast(o, program_ctx)
  elif tf_inspect.ismethod(o):
    nodes, name, entity_info = convert_func_to_ast(o, program_ctx)
  elif hasattr(o, '__class__'):
    # Note: this should only be raised when attempting to convert the object
    # directly. converted_call should still support it.
    raise NotImplementedError(
        'cannot convert entity "{}": object conversion is not yet'
        ' supported.'.format(o))
  else:
    raise ValueError(
        'Entity "%s" has unsupported type "%s". Only functions and classes are '
        'supported for now.' % (o, type(o)))

  if logging.has_verbosity(2):
    logging.log(2, 'Compiled output of %s:\n\n%s\n', o,
                compiler.ast_to_source(nodes))
  if logging.has_verbosity(4):
    for n in nodes:
      logging.log(4, 'Compiled AST of %s:\n\n%s\n\n', o,
                  pretty_printer.fmt(n, color=False))

  return nodes, name, entity_info


def convert_class_to_ast(c, program_ctx):
  """Specialization of `convert_entity_to_ast` for classes."""
  # TODO(mdan): Revisit this altogether. Not sure we still need it.
  converted_members = {}
  method_filter = lambda m: tf_inspect.isfunction(m) or tf_inspect.ismethod(m)
  members = tf_inspect.getmembers(c, predicate=method_filter)
  if not members:
    raise ValueError('cannot convert %s: no member methods' % c)

  # TODO(mdan): Don't clobber namespaces for each method in one class namespace.
  # The assumption that one namespace suffices for all methods only holds if
  # all methods were defined in the same module.
  # If, instead, functions are imported from multiple modules and then spliced
  # into the class, then each function has its own globals and __future__
  # imports that need to stay separate.

  # For example, C's methods could both have `global x` statements referring to
  # mod1.x and mod2.x, but using one namespace for C would cause a conflict.
  # from mod1 import f1
  # from mod2 import f2
  # class C(object):
  #   method1 = f1
  #   method2 = f2

  class_namespace = {}
  future_features = None
  for _, m in members:
    # Only convert the members that are directly defined by the class.
    if inspect_utils.getdefiningclass(m, c) is not c:
      continue
    (node,), _, entity_info = convert_func_to_ast(
        m,
        program_ctx=program_ctx,
        do_rename=False)
    class_namespace.update(entity_info.namespace)
    converted_members[m] = node

    # TODO(mdan): Similarly check the globals.
    if future_features is None:
      future_features = entity_info.future_features
    elif frozenset(future_features) ^ frozenset(entity_info.future_features):
      # Note: we can support this case if ever needed.
      raise ValueError(
          'cannot convert {}: if has methods built with mismatched future'
          ' features: {} and {}'.format(c, future_features,
                                        entity_info.future_features))
  namer = naming.Namer(class_namespace)
  class_name = namer.class_name(c.__name__)

  # Process any base classes: if the superclass if of a whitelisted type, an
  # absolute import line is generated.
  output_nodes = []
  renames = {}
  base_names = []
  for base in c.__bases__:
    if isinstance(object, base):
      base_names.append('object')
      continue
    if is_whitelisted_for_graph(base):
      alias = namer.new_symbol(base.__name__, ())
      output_nodes.append(
          gast.ImportFrom(
              module=base.__module__,
              names=[gast.alias(name=base.__name__, asname=alias)],
              level=0))
    else:
      raise NotImplementedError(
          'Conversion of classes that do not directly extend classes from'
          ' whitelisted modules is temporarily suspended. If this breaks'
          ' existing code please notify the AutoGraph team immediately.')
    base_names.append(alias)
    renames[qual_names.QN(base.__name__)] = qual_names.QN(alias)

  # Generate the definition of the converted class.
  bases = [gast.Name(n, gast.Load(), None) for n in base_names]
  class_def = gast.ClassDef(
      class_name,
      bases=bases,
      keywords=[],
      body=list(converted_members.values()),
      decorator_list=[])
  # Make a final pass to replace references to the class or its base classes.
  # Most commonly, this occurs when making super().__init__() calls.
  # TODO(mdan): Making direct references to superclass' superclass will fail.
  class_def = qual_names.resolve(class_def)
  renames[qual_names.QN(c.__name__)] = qual_names.QN(class_name)
  class_def = ast_util.rename_symbols(class_def, renames)

  output_nodes.append(class_def)

  # TODO(mdan): Find a way better than forging this object.
  entity_info = transformer.EntityInfo(
      source_code=None,
      source_file=None,
      future_features=future_features,
      namespace=class_namespace)

  return output_nodes, class_name, entity_info


def _add_reserved_symbol(namespace, name, entity):
  if name not in namespace:
    namespace[name] = entity
  elif namespace[name] != entity:
    raise ValueError('The name "%s" is reserved and may not be used.' % name)


ag_internal = None


# TODO(mdan): Move into core or replace with an actual importable module.
def _add_self_references(namespace, autograph_module):
  """Adds namespace references to the module that exposes the api itself."""
  global ag_internal
  if ag_internal is None:
    # Craft a module that exposes parts of the external API as well as certain
    # internal modules.
    ag_internal = imp.new_module('autograph')
    ag_internal.__dict__.update(autograph_module.__dict__)
    ag_internal.ConversionOptions = converter.ConversionOptions
    ag_internal.Feature = converter.Feature
    ag_internal.utils = utils
    ag_internal.function_scope = function_wrapping.function_scope
    # TODO(mdan): Add safeguards against name clashes.
    # We don't want to create a submodule because we want the operators to be
    # accessible as ag__.<operator>
    ag_internal.__dict__.update(special_functions.__dict__)
    ag_internal.__dict__.update(operators.__dict__)

  _add_reserved_symbol(namespace, 'ag__', ag_internal)


def convert_func_to_ast(f, program_ctx, do_rename=True):
  """Specialization of `convert_entity_to_ast` for callable functions."""

  future_features = inspect_utils.getfutureimports(f)
  node, source = parser.parse_entity(f, future_features=future_features)
  logging.log(3, 'Source code of %s:\n\n%s\n', f, source)
  # Parsed AST should contain future imports and one function def node.

  # In general, the output of inspect.getsource is inexact for lambdas because
  # it uses regex matching to adjust the exact location around the line number
  # that CPython records. Then, the entire containing line is returned, which
  # we may have trouble disambiguating. For example:
  # x, y = lambda: 1, lambda: 2
  if f.__name__ == '<lambda>':
    nodes = ast_util.find_matching_definitions(node, f)
    if len(nodes) != 1:
      raise ValueError(
          'Unable to identify source code of lambda function {}. It was'
          ' defined on this line: {}, which must contain a single lambda with'
          ' matching signature. To avoid ambiguity, define each lambda'
          ' in a separate expression.'.format(f, source))
    node, = nodes

  # TODO(znado): Place inside standard_analysis.
  origin_info.resolve_entity(node, source, f)

  namespace = inspect_utils.getnamespace(f)
  _add_self_references(namespace, program_ctx.autograph_module)
  namer = naming.Namer(namespace)

  entity_info = transformer.EntityInfo(
      source_code=source,
      source_file='<fragment>',
      future_features=future_features,
      namespace=namespace)
  context = converter.EntityContext(namer, entity_info, program_ctx)
  node = node_to_graph(node, context)

  if isinstance(node, gast.Lambda):
    new_name = namer.new_symbol('tf__lambda', ())
    node = gast.Assign(
        targets=[gast.Name(new_name, gast.Store(), None)], value=node)

  elif do_rename:
    new_name = namer.function_name(f.__name__)
    node.name = new_name
  else:
    new_name = f.__name__
    assert node.name == new_name

  return (node,), new_name, entity_info


def node_to_graph(node, context):
  """Convert Python code to equivalent TF graph mode code.

  Args:
    node: AST, the code to convert.
    context: converter.EntityContext

  Returns:
    A tuple (node, deps):
        * node: A Python ast node, representing the converted code.
        * deps: A set of strings, the fully qualified names of entity
            dependencies that this node has.
  """
  # TODO(mdan): Insert list_comprehensions somewhere.
  unsupported_features_checker.verify(node)

  node = converter.standard_analysis(node, context, is_initial=True)
  node = converter.apply_(node, context, arg_defaults)
  node = converter.apply_(node, context, directives)
  node = converter.apply_(node, context, break_statements)
  if context.program.options.uses(converter.Feature.ASSERT_STATEMENTS):
    node = converter.apply_(node, context, asserts)
  # Note: sequencing continue canonicalization before for loop one avoids
  # dealing with the extra loop increment operation that the for
  # canonicalization creates.
  node = converter.apply_(node, context, continue_statements)
  node = converter.apply_(node, context, return_statements)
  if context.program.options.uses(converter.Feature.LISTS):
    node = converter.apply_(node, context, lists)
    node = converter.apply_(node, context, slices)
  node = converter.apply_(node, context, call_trees)
  node = converter.apply_(node, context, control_flow)
  node = converter.apply_(node, context, conditional_expressions)
  if context.program.options.uses(converter.Feature.LOGICAL_EXPRESSIONS):
    node = converter.apply_(node, context, logical_expressions)
  if context.program.options.uses(converter.Feature.AUTO_CONTROL_DEPS):
    node = converter.apply_(node, context, side_effect_guards)
  # TODO(mdan): If function scopes ever does more, the toggle will need moving.
  if context.program.options.uses(converter.Feature.NAME_SCOPES):
    node = converter.apply_(node, context, function_scopes)
  return node



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A TF-GAN-backed StarGAN Estimator."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import enum

from tensorflow.contrib.framework.python.ops import variables as variable_lib
from tensorflow.contrib.gan.python import namedtuples as tfgan_tuples
from tensorflow.contrib.gan.python import train as tfgan_train
from tensorflow.contrib.gan.python.eval.python import summaries as tfgan_summaries
from tensorflow.python.estimator import estimator
from tensorflow.python.estimator import model_fn as model_fn_lib
from tensorflow.python.framework import ops
from tensorflow.python.ops import metrics as metrics_lib
from tensorflow.python.ops import variable_scope
from tensorflow.python.util import tf_inspect as inspect

__all__ = ['StarGANEstimator', 'SummaryType']


class SummaryType(enum.IntEnum):
  NONE = 0
  VARIABLES = 1
  IMAGES = 2
  IMAGE_COMPARISON = 3


_summary_type_map = {
    SummaryType.VARIABLES: tfgan_summaries.add_gan_model_summaries,
    SummaryType.IMAGES: tfgan_summaries.add_stargan_image_summaries,
}


class StarGANEstimator(estimator.Estimator):
  """An estimator for Generative Adversarial Networks (GANs).

  This Estimator is backed by TFGAN. The network functions follow the TFGAN API
  except for one exception: if either `generator_fn` or `discriminator_fn` have
  an argument called `mode`, then the tf.Estimator mode is passed in for that
  argument. This helps with operations like batch normalization, which have
  different train and evaluation behavior.

  Example:

  ```python
      import tensorflow as tf
      tfgan = tf.contrib.gan

      # See TFGAN's `train.py` for a description of the generator and
      # discriminator API.
      def generator_fn(generator_inputs):
        ...
        return generated_data

      def discriminator_fn(data, conditioning):
        ...
        return logits

      # Create GAN estimator.
      stargan_estimator = tfgan.estimator.StarGANEstimator(
          model_dir,
          generator_fn=generator_fn,
          discriminator_fn=discriminator_fn,
          loss_fn=loss_fn,
          generator_optimizer=tf.compat.v1.train.AdamOptimizer(0.1, 0.5),
          discriminator_optimizer=tf.compat.v1.train.AdamOptimizer(0.1, 0.5))

      # Train estimator.
      stargan_estimator.train(train_input_fn, steps)

      # Evaluate resulting estimator.
      stargan_estimator.evaluate(eval_input_fn)

      # Generate samples from generator.
      stargan_estimator = np.array([
          x for x in stargan_estimator.predict(predict_input_fn)])
  ```
  """

  def __init__(self,
               model_dir=None,
               generator_fn=None,
               discriminator_fn=None,
               loss_fn=None,
               generator_optimizer=None,
               discriminator_optimizer=None,
               get_hooks_fn=None,
               get_eval_metric_ops_fn=None,
               add_summaries=None,
               use_loss_summaries=True,
               config=None):
    """Initializes a StarGANEstimator instance.

    Args:
      model_dir: Directory to save model parameters, graph and etc. This can
        also be used to load checkpoints from the directory into a estimator to
        continue training a previously saved model.
      generator_fn: A python function that takes a Tensor, Tensor list, or
        Tensor dictionary as inputs and returns the outputs of the GAN
        generator. See `TFGAN` for more details and examples. Additionally, if
        it has an argument called `mode`, the Estimator's `mode` will be passed
        in (ex TRAIN, EVAL, PREDICT). This is useful for things like batch
        normalization.
      discriminator_fn: A python function that takes the output of
        `generator_fn` or real data in the GAN setup, and `input_data`. Outputs
        a Tensor in the range [-inf, inf]. See `TFGAN` for more details and
        examples.
      loss_fn: The loss function on the generator. Takes a `StarGANModel`
        namedtuple and return a `GANLoss` namedtuple.
      generator_optimizer: The optimizer for generator updates, or a function
        that takes no arguments and returns an optimizer. This function will be
        called when the default graph is the `StarGANEstimator`'s graph, so
        utilities like `tf.contrib.framework.get_or_create_global_step` will
        work.
      discriminator_optimizer: Same as `generator_optimizer`, but for the
        discriminator updates.
      get_hooks_fn: A function that takes a `GANTrainOps` tuple and returns a
        list of hooks. These hooks are run on the generator and discriminator
        train ops, and can be used to implement the GAN training scheme.
        Defaults to `train.get_sequential_train_hooks()`.
      get_eval_metric_ops_fn: A function that takes a `GANModel`, and returns a
        dict of metric results keyed by name. The output of this function is
        passed into `tf.estimator.EstimatorSpec` during evaluation.
      add_summaries: `None`, a single `SummaryType`, or a list of `SummaryType`.
      use_loss_summaries: If `True`, add loss summaries. If `False`, does not.
        If `None`, uses defaults.
      config: `RunConfig` object to configure the runtime settings.

    Raises:
      ValueError: If loss functions aren't callable.
      ValueError: If `use_loss_summaries` isn't boolean or `None`.
      ValueError: If `get_hooks_fn` isn't callable or `None`.
    """
    if not callable(loss_fn):
      raise ValueError('loss_fn must be callable.')
    if use_loss_summaries not in [True, False, None]:
      raise ValueError('use_loss_summaries must be True, False or None.')
    if get_hooks_fn is not None and not callable(get_hooks_fn):
      raise TypeError('get_hooks_fn must be callable.')

    def _model_fn(features, labels, mode):
      """StarGANEstimator model function."""
      if mode not in [
          model_fn_lib.ModeKeys.TRAIN, model_fn_lib.ModeKeys.EVAL,
          model_fn_lib.ModeKeys.PREDICT
      ]:
        raise ValueError('Mode not recognized: %s' % mode)

      if mode == model_fn_lib.ModeKeys.PREDICT:
        input_data = features[0]
        input_data_domain_label = features[1]
      else:
        input_data = features  # rename inputs for clarity
        input_data_domain_label = labels  # rename inputs for clarity

      # Make StarGANModel, which encapsulates the GAN model architectures.
      gan_model = _get_gan_model(mode, generator_fn, discriminator_fn,
                                 input_data, input_data_domain_label,
                                 add_summaries)

      # Make the EstimatorSpec, which incorporates the StarGANModel, losses,
      # eval, metrics, and optimizers (if required).
      return _get_estimator_spec(mode, gan_model, loss_fn,
                                 get_eval_metric_ops_fn, generator_optimizer,
                                 discriminator_optimizer, get_hooks_fn)

    super(StarGANEstimator, self).__init__(
        model_fn=_model_fn, model_dir=model_dir, config=config)


def _get_gan_model(mode,
                   generator_fn,
                   discriminator_fn,
                   input_data,
                   input_data_domain_label,
                   add_summaries,
                   generator_scope='Generator'):
  """Makes the StarGANModel tuple."""
  if mode == model_fn_lib.ModeKeys.PREDICT:
    gan_model = _make_prediction_gan_model(input_data, input_data_domain_label,
                                           generator_fn, generator_scope)
  else:  # model_fn_lib.ModeKeys.TRAIN or model_fn_lib.ModeKeys.EVAL
    gan_model = _make_gan_model(generator_fn, discriminator_fn, input_data,
                                input_data_domain_label, generator_scope,
                                add_summaries, mode)

  return gan_model


def _get_estimator_spec(mode,
                        gan_model,
                        loss_fn,
                        get_eval_metric_ops_fn,
                        generator_optimizer,
                        discriminator_optimizer,
                        get_hooks_fn=None):
  """Get the EstimatorSpec for the current mode."""
  if mode == model_fn_lib.ModeKeys.PREDICT:
    estimator_spec = model_fn_lib.EstimatorSpec(
        mode=mode, predictions=gan_model.generated_data)
  else:
    gan_loss = loss_fn(gan_model)
    if mode == model_fn_lib.ModeKeys.EVAL:
      estimator_spec = _get_eval_estimator_spec(gan_model, gan_loss,
                                                get_eval_metric_ops_fn)
    else:  # model_fn_lib.ModeKeys.TRAIN:
      gopt = (
          generator_optimizer()
          if callable(generator_optimizer) else generator_optimizer)
      dopt = (
          discriminator_optimizer()
          if callable(discriminator_optimizer) else discriminator_optimizer)
      get_hooks_fn = get_hooks_fn or tfgan_train.get_sequential_train_hooks()
      estimator_spec = _get_train_estimator_spec(gan_model, gan_loss, gopt,
                                                 dopt, get_hooks_fn)

  return estimator_spec


def _make_gan_model(generator_fn, discriminator_fn, input_data,
                    input_data_domain_label, generator_scope, add_summaries,
                    mode):
  """Construct a `StarGANModel`, and optionally pass in `mode`."""
  # If network functions have an argument `mode`, pass mode to it.
  if 'mode' in inspect.getargspec(generator_fn).args:
    generator_fn = functools.partial(generator_fn, mode=mode)
  if 'mode' in inspect.getargspec(discriminator_fn).args:
    discriminator_fn = functools.partial(discriminator_fn, mode=mode)
  gan_model = tfgan_train.stargan_model(
      generator_fn,
      discriminator_fn,
      input_data,
      input_data_domain_label,
      generator_scope=generator_scope)
  if add_summaries:
    if not isinstance(add_summaries, (tuple, list)):
      add_summaries = [add_summaries]
    with ops.name_scope(None):
      for summary_type in add_summaries:
        _summary_type_map[summary_type](gan_model)

  return gan_model


def _make_prediction_gan_model(input_data, input_data_domain_label,
                               generator_fn, generator_scope):
  """Make a `StarGANModel` from just the generator."""
  # If `generator_fn` has an argument `mode`, pass mode to it.
  if 'mode' in inspect.getargspec(generator_fn).args:
    generator_fn = functools.partial(
        generator_fn, mode=model_fn_lib.ModeKeys.PREDICT)
  with variable_scope.variable_scope(generator_scope) as gen_scope:
    # pylint:disable=protected-access
    input_data = tfgan_train._convert_tensor_or_l_or_d(input_data)
    input_data_domain_label = tfgan_train._convert_tensor_or_l_or_d(
        input_data_domain_label)
    # pylint:enable=protected-access
    generated_data = generator_fn(input_data, input_data_domain_label)
  generator_variables = variable_lib.get_trainable_variables(gen_scope)

  return tfgan_tuples.StarGANModel(
      input_data=input_data,
      input_data_domain_label=None,
      generated_data=generated_data,
      generated_data_domain_target=input_data_domain_label,
      reconstructed_data=None,
      discriminator_input_data_source_predication=None,
      discriminator_generated_data_source_predication=None,
      discriminator_input_data_domain_predication=None,
      discriminator_generated_data_domain_predication=None,
      generator_variables=generator_variables,
      generator_scope=generator_scope,
      generator_fn=generator_fn,
      discriminator_variables=None,
      discriminator_scope=None,
      discriminator_fn=None)


def _get_eval_estimator_spec(gan_model,
                             gan_loss,
                             get_eval_metric_ops_fn=None,
                             name=None):
  """Return an EstimatorSpec for the eval case."""
  scalar_loss = gan_loss.generator_loss + gan_loss.discriminator_loss
  with ops.name_scope(None, 'metrics',
                      [gan_loss.generator_loss, gan_loss.discriminator_loss]):

    def _summary_key(head_name, val):
      return '%s/%s' % (val, head_name) if head_name else val

    eval_metric_ops = {
        _summary_key(name, 'generator_loss'):
            metrics_lib.mean(gan_loss.generator_loss),
        _summary_key(name, 'discriminator_loss'):
            metrics_lib.mean(gan_loss.discriminator_loss)
    }
    if get_eval_metric_ops_fn is not None:
      custom_eval_metric_ops = get_eval_metric_ops_fn(gan_model)
      if not isinstance(custom_eval_metric_ops, dict):
        raise TypeError('get_eval_metric_ops_fn must return a dict, '
                        'received: {}'.format(custom_eval_metric_ops))
      eval_metric_ops.update(custom_eval_metric_ops)
  return model_fn_lib.EstimatorSpec(
      mode=model_fn_lib.ModeKeys.EVAL,
      predictions=gan_model.generated_data,
      loss=scalar_loss,
      eval_metric_ops=eval_metric_ops)


def _get_train_estimator_spec(gan_model,
                              gan_loss,
                              generator_optimizer,
                              discriminator_optimizer,
                              get_hooks_fn,
                              train_op_fn=tfgan_train.gan_train_ops):
  """Return an EstimatorSpec for the train case."""
  scalar_loss = gan_loss.generator_loss + gan_loss.discriminator_loss
  train_ops = train_op_fn(gan_model, gan_loss, generator_optimizer,
                          discriminator_optimizer)
  training_hooks = get_hooks_fn(train_ops)
  return model_fn_lib.EstimatorSpec(
      loss=scalar_loss,
      mode=model_fn_lib.ModeKeys.TRAIN,
      train_op=train_ops.global_step_inc_op,
      training_hooks=training_hooks)


def stargan_prediction_input_fn_wrapper(fn):
  """StarGAN Estimator prediction input_fn wrapper.

  Since estimator will disregard the "label" variable pass to the model, we will
  use a wrapper to pack the (feature, label) tuple as feature passed to the
  model.

  Args:
    fn: input_fn for the prediction.

  Returns:
    A tuple ((feature, label), None) where the second element is the dummy label
    to be disregarded and the first element is the true input to the estimator.
  """

  def new_fn():
    return fn(), None

  return new_fn



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The Exponential distribution class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn
from tensorflow.python.ops import random_ops
from tensorflow.python.ops.distributions import gamma
from tensorflow.python.util import deprecation
from tensorflow.python.util.tf_export import tf_export


__all__ = [
    "Exponential",
    "ExponentialWithSoftplusRate",
]


@tf_export(v1=["distributions.Exponential"])
class Exponential(gamma.Gamma):
  """Exponential distribution.

  The Exponential distribution is parameterized by an event `rate` parameter.

  #### Mathematical Details

  The probability density function (pdf) is,

  ```none
  pdf(x; lambda, x > 0) = exp(-lambda x) / Z
  Z = 1 / lambda
  ```

  where `rate = lambda` and `Z` is the normalizaing constant.

  The Exponential distribution is a special case of the Gamma distribution,
  i.e.,

  ```python
  Exponential(rate) = Gamma(concentration=1., rate)
  ```

  The Exponential distribution uses a `rate` parameter, or "inverse scale",
  which can be intuited as,

  ```none
  X ~ Exponential(rate=1)
  Y = X / rate
  ```

  """

  @deprecation.deprecated(
      "2019-01-01",
      "The TensorFlow Distributions library has moved to "
      "TensorFlow Probability "
      "(https://github.com/tensorflow/probability). You "
      "should update all references to use `tfp.distributions` "
      "instead of `tf.distributions`.",
      warn_once=True)
  def __init__(self,
               rate,
               validate_args=False,
               allow_nan_stats=True,
               name="Exponential"):
    """Construct Exponential distribution with parameter `rate`.

    Args:
      rate: Floating point tensor, equivalent to `1 / mean`. Must contain only
        positive values.
      validate_args: Python `bool`, default `False`. When `True` distribution
        parameters are checked for validity despite possibly degrading runtime
        performance. When `False` invalid inputs may silently render incorrect
        outputs.
      allow_nan_stats: Python `bool`, default `True`. When `True`, statistics
        (e.g., mean, mode, variance) use the value "`NaN`" to indicate the
        result is undefined. When `False`, an exception is raised if one or
        more of the statistic's batch members are undefined.
      name: Python `str` name prefixed to Ops created by this class.
    """
    parameters = dict(locals())
    # Even though all statistics of are defined for valid inputs, this is not
    # true in the parent class "Gamma."  Therefore, passing
    # allow_nan_stats=True
    # through to the parent class results in unnecessary asserts.
    with ops.name_scope(name, values=[rate]) as name:
      self._rate = ops.convert_to_tensor(rate, name="rate")
    super(Exponential, self).__init__(
        concentration=array_ops.ones([], dtype=self._rate.dtype),
        rate=self._rate,
        allow_nan_stats=allow_nan_stats,
        validate_args=validate_args,
        name=name)
    self._parameters = parameters
    self._graph_parents += [self._rate]

  @staticmethod
  def _param_shapes(sample_shape):
    return {"rate": ops.convert_to_tensor(sample_shape, dtype=dtypes.int32)}

  @property
  def rate(self):
    return self._rate

  def _log_survival_function(self, value):
    return self._log_prob(value) - math_ops.log(self._rate)

  def _sample_n(self, n, seed=None):
    shape = array_ops.concat([[n], array_ops.shape(self._rate)], 0)
    # Uniform variates must be sampled from the open-interval `(0, 1)` rather
    # than `[0, 1)`. To do so, we use `np.finfo(self.dtype.as_numpy_dtype).tiny`
    # because it is the smallest, positive, "normal" number. A "normal" number
    # is such that the mantissa has an implicit leading 1. Normal, positive
    # numbers x, y have the reasonable property that, `x + y >= max(x, y)`. In
    # this case, a subnormal number (i.e., np.nextafter) can cause us to sample
    # 0.
    sampled = random_ops.random_uniform(
        shape,
        minval=np.finfo(self.dtype.as_numpy_dtype).tiny,
        maxval=1.,
        seed=seed,
        dtype=self.dtype)
    return -math_ops.log(sampled) / self._rate


class ExponentialWithSoftplusRate(Exponential):
  """Exponential with softplus transform on `rate`."""

  @deprecation.deprecated(
      "2019-01-01",
      "Use `tfd.Exponential(tf.nn.softplus(rate)).",
      warn_once=True)
  def __init__(self,
               rate,
               validate_args=False,
               allow_nan_stats=True,
               name="ExponentialWithSoftplusRate"):
    parameters = dict(locals())
    with ops.name_scope(name, values=[rate]) as name:
      super(ExponentialWithSoftplusRate, self).__init__(
          rate=nn.softplus(rate, name="softplus_rate"),
          validate_args=validate_args,
          allow_nan_stats=allow_nan_stats,
          name=name)
    self._parameters = parameters



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Ops for tensor_forest."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python import ops
from tensorflow.python.ops import gen_tensor_forest_ops
from tensorflow.python.ops import resources
from tensorflow.python.training import saver


class TreeVariableSaveable(saver.BaseSaverBuilder.SaveableObject):
  """Resource that holds a tree."""

  def __init__(self, type_name, name, container, config, resource_handle_func,
               create_op_func, is_initialized_op_func, serialize_op_func,
               deserialize_op_func):

    with ops.name_scope(name, type_name) as name:
      self._resource_handle = resource_handle_func(
          container, shared_name=name, name=name)

    self._is_initialized_op = is_initialized_op_func(self._resource_handle)
    tensor = serialize_op_func(self._resource_handle)
    self._create_op = create_op_func(self._resource_handle, config)
    # slice_spec is useful for saving a slice from a variable.
    # It's not meaningful the tree variable. So we just pass an empty
    # value.
    slice_spec = ''
    specs = [saver.BaseSaverBuilder.SaveSpec(tensor, slice_spec, name)]
    super(TreeVariableSaveable, self).__init__(self._resource_handle, specs,
                                               name)

    ops.add_to_collection(ops.GraphKeys.SAVEABLE_OBJECTS, self)

    resources.register_resource(self._resource_handle, self._create_op,
                                self._is_initialized_op)
    self._deserialize_op_func = deserialize_op_func

  def restore(self, restored_tensors, unused_restored_shapes):
    """Restores the associated tree from 'restored_tensors'.

    Args:
      restored_tensors: the tensors that were loaded from a checkpoint.
      unused_restored_shapes: the shapes this object should conform to after
        restore. Not meaningful for trees.

    Returns:
      The operation that restores the state of the tree variable.
    """
    with ops.control_dependencies([self._create_op]):
      return self._deserialize_op_func(
          self._resource_handle,
          restored_tensors[0],
      )

  @property
  def resource(self):
    return self._resource_handle


def tree_variable(tree_config, name, container=None):
  return TreeVariableSaveable(
      'TreeVariable', name, container, tree_config,
      gen_tensor_forest_ops.tensor_forest_tree_resource_handle_op,
      gen_tensor_forest_ops.tensor_forest_create_tree_variable,
      gen_tensor_forest_ops.tensor_forest_tree_is_initialized_op,
      gen_tensor_forest_ops.tensor_forest_tree_serialize,
      gen_tensor_forest_ops.tensor_forest_tree_deserialize).resource


class ForestVariables(object):
  """Resource that holds all trees from a forest."""

  def __init__(self, params, tree_configs=None):

    self._variables = []

    for i in range(params.n_trees):
      tree_config = ''
      if tree_configs is not None:
        tree_config = tree_configs[i]
      self._variables.append(tree_variable(
          tree_config,
          'tree-%s' % i,
      ))

  def __getitem__(self, t):
    return self._variables[t]
